# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, TYPE_CHECKING

from iconcommons import Logger

from ...base.exception import FatalException
from ...icon_constant import ISCORE_EXCHANGE_RATE, IISS_LOG_TAG

if TYPE_CHECKING:
    from .storage import RegulatorVariable
    from ...iconscore.icon_score_context import IconScoreContext


class Regulator:
    """ Regulate ICX issue amount
    * The difference of I-SCORE calculation between icon service and reward calculator
    * Fee from the transaction

    covered_icx_by_fee:
    Used for generating coin base transaction and emitting event log

    covered_icx_by_over_issue:
    Used for generating coin base transaction and emitting event log. It could be
    negative value (In case that Reward calculator issued ICX more then Icon service)

    remain_over_issued_icx:
    Only used for emitting eventlog

    corrected_icx_issue_amount:
    Used for issuing, generating coin base transaction and emitting event log
    """
    def __init__(self, context: 'IconScoreContext', issue_amount: int):
        self._regulator_variable: Optional['RegulatorVariable'] = None
        self._covered_icx_by_fee: Optional[int] = None
        self._covered_icx_by_remain: Optional[int] = None
        self._corrected_icx_issue_amount: Optional[int] = None

        self._set_corrected_issue_data(context, issue_amount)

    @property
    def covered_icx_by_fee(self) -> int:
        return self._covered_icx_by_fee

    @property
    def covered_icx_by_over_issue(self) -> int:
        return self._covered_icx_by_remain

    @property
    def remain_over_issued_icx(self) -> int:
        return self._regulator_variable.over_issued_iscore // ISCORE_EXCHANGE_RATE

    @property
    def corrected_icx_issue_amount(self) -> int:
        return self._corrected_icx_issue_amount

    def _set_corrected_issue_data(self, context: 'IconScoreContext', issue_amount: int):
        regulator_variable: 'RegulatorVariable' = context.storage.issue.get_regulator_variable(context)
        prev_block_cumulative_fee: int = context.storage.icx.last_block.cumulative_fee
        end_block_height_of_calc: int = context.storage.iiss.get_end_block_height_of_calc(context)

        # Update current calculated period total issued icx
        current_calc_period_total_issued_icx: int = regulator_variable.current_calc_period_issued_icx
        current_calc_period_total_issued_icx += issue_amount
        if end_block_height_of_calc == context.block.height:
            prev_calc_period_issued_iscore, _, _ = context.storage.rc.get_calc_response_from_rc()

            assert prev_calc_period_issued_iscore >= 0

            # In case of the first term of decentralization.
            # Do not regulate on the first term of decentralization
            # as Icon service has not issued ICX on the last period of 'pre-vote'
            # (On pre-vote, icon-foundation provided ICX instead of issuing it)
            if regulator_variable.prev_calc_period_issued_icx == -1:
                regulator_variable.prev_calc_period_issued_icx, prev_calc_period_issued_iscore = 0, 0
            covered_icx_by_fee, covered_icx_by_remain, remain_over_issued_iscore, corrected_icx_issue_amount = \
                self._correct_issue_amount_on_calc_period(regulator_variable.prev_calc_period_issued_icx,
                                                          prev_calc_period_issued_iscore,
                                                          regulator_variable.over_issued_iscore,
                                                          issue_amount,
                                                          prev_block_cumulative_fee)

            regulator_variable.prev_calc_period_issued_icx = current_calc_period_total_issued_icx
            regulator_variable.current_calc_period_issued_icx = 0
        else:
            covered_icx_by_fee, covered_icx_by_remain, remain_over_issued_iscore, corrected_icx_issue_amount = \
                self._correct_issue_amount(regulator_variable.over_issued_iscore,
                                           issue_amount,
                                           prev_block_cumulative_fee)
            regulator_variable.current_calc_period_issued_icx = current_calc_period_total_issued_icx
        regulator_variable.over_issued_iscore = remain_over_issued_iscore

        self._regulator_variable = regulator_variable
        self._covered_icx_by_fee = covered_icx_by_fee
        self._covered_icx_by_remain = covered_icx_by_remain
        self._corrected_icx_issue_amount = corrected_icx_issue_amount
        Logger.info(f"Regulate BH: {context.block.height} "
                    f"Covered by fee: {self._covered_icx_by_fee} "
                    f"Covered by remain: {self._covered_icx_by_remain} "
                    f"Corrected issue amount {self._corrected_icx_issue_amount}"
                    f"Regulator variable: {self._regulator_variable}", IISS_LOG_TAG)

    def put_regulate_variable(self, context: 'IconScoreContext'):
        context.storage.issue.put_regulator_variable(context, self._regulator_variable)

    @classmethod
    def _reflect_difference_in_issuing(cls,
                                       icx_issue_amount: int,
                                       over_issued_icx: int,
                                       prev_block_cumulative_fee: int) -> Tuple[int, int, int, int]:
        # cover about over issued icx
        # remain_over_issued_icx, corrected_icx_issue_amount
        covered_icx_by_remain, remain_over_issued_icx, corrected_icx_issue_amount = \
            cls._calculate_over_issued_icx(icx_issue_amount,
                                           over_issued_icx)

        covered_icx_by_fee, remain_over_issued_icx, corrected_icx_issue_amount = \
            cls._calculate_prev_block_cumulative_fee(remain_over_issued_icx,
                                                     corrected_icx_issue_amount,
                                                     prev_block_cumulative_fee)

        return covered_icx_by_fee, covered_icx_by_remain, remain_over_issued_icx, corrected_icx_issue_amount

    @classmethod
    def _calculate_over_issued_icx(cls,
                                   icx_issue_amount: int,
                                   over_issued_icx: int) -> Tuple[int, int, int]:
        corrected_issue_amount = icx_issue_amount - over_issued_icx
        # In case that Reward calculator issued ICX more then Icon service
        if over_issued_icx < 0:
            remain_over_issued_icx = 0
            covered_icx_by_remain = over_issued_icx
            return covered_icx_by_remain, remain_over_issued_icx, corrected_issue_amount

        if corrected_issue_amount >= 0:
            remain_over_issued_icx = 0
            covered_icx_by_remain = over_issued_icx
        else:
            remain_over_issued_icx = abs(corrected_issue_amount)
            corrected_issue_amount = 0
            covered_icx_by_remain = icx_issue_amount

        return covered_icx_by_remain, remain_over_issued_icx, corrected_issue_amount

    @classmethod
    def _calculate_prev_block_cumulative_fee(cls,
                                             remain_over_issued_icx: int,
                                             corrected_issue_amount: int,
                                             prev_block_cumulative_fee: int) -> Tuple[int, int, int]:
        corrected_issue_amount -= prev_block_cumulative_fee
        if corrected_issue_amount >= 0:
            covered_icx_by_fee = prev_block_cumulative_fee
        else:
            covered_icx_by_fee = prev_block_cumulative_fee + corrected_issue_amount
            remain_over_issued_icx = remain_over_issued_icx + abs(corrected_issue_amount)
            corrected_issue_amount = 0

        return covered_icx_by_fee, remain_over_issued_icx, corrected_issue_amount

    @classmethod
    def _calculate_prev_block_cumulative_fee(cls,
                                             remain_over_issued_icx: int,
                                             corrected_issue_amount: int,
                                             prev_block_cumulative_fee: int) -> Tuple[int, int, int]:
        corrected_issue_amount -= prev_block_cumulative_fee
        if corrected_issue_amount >= 0:
            covered_icx_by_fee = prev_block_cumulative_fee
        else:
            covered_icx_by_fee = prev_block_cumulative_fee + corrected_issue_amount
            remain_over_issued_icx = remain_over_issued_icx + abs(corrected_issue_amount)
            corrected_issue_amount = 0

        return covered_icx_by_fee, remain_over_issued_icx, corrected_issue_amount

    @classmethod
    def _separate_icx_and_iscore(cls, iscore: int) -> Tuple[int, int]:
        abs_iscore = abs(iscore)
        over_issued_icx = abs_iscore // ISCORE_EXCHANGE_RATE
        over_issued_iscore = abs_iscore % ISCORE_EXCHANGE_RATE
        if iscore < 0:
            over_issued_icx = -over_issued_icx
            over_issued_iscore = -over_issued_iscore

        return over_issued_icx, over_issued_iscore

    @classmethod
    def _correct_issue_amount_on_calc_period(cls,
                                             prev_calc_period_issued_icx: int,
                                             prev_calc_period_issued_iscore: int,
                                             remain_over_issued_iscore: int,
                                             icx_issue_amount: int,
                                             prev_block_cumulative_fee: int) -> Tuple[int, int, int, int]:
        """

        :param prev_calc_period_issued_icx: Calculated from Icon Service
        :param prev_calc_period_issued_iscore: Calculated from Reward Calculator
        :param remain_over_issued_iscore: Amount
        :param icx_issue_amount:
        :param prev_block_cumulative_fee:
        :return:
        """
        assert icx_issue_amount >= 0
        assert prev_block_cumulative_fee >= 0
        assert prev_calc_period_issued_iscore >= 0

        prev_calc_over_issued_iscore: int = \
            prev_calc_period_issued_icx * ISCORE_EXCHANGE_RATE - prev_calc_period_issued_iscore
        total_over_issued_iscore: int = prev_calc_over_issued_iscore + remain_over_issued_iscore
        # Over issued amount could be negative value. It means reward calculator issued more then icon service.
        over_issued_icx, over_issued_iscore = cls._separate_icx_and_iscore(total_over_issued_iscore)

        covered_icx_by_fee, covered_icx_by_remain, remain_over_issued_icx, icx_issue_amount = \
            cls._reflect_difference_in_issuing(icx_issue_amount, over_issued_icx, prev_block_cumulative_fee)

        remain_over_issued_iscore: int = remain_over_issued_icx * ISCORE_EXCHANGE_RATE + over_issued_iscore

        # covered_icx can be negative value (in case of reward calculator having been issued more)
        return covered_icx_by_fee, covered_icx_by_remain, remain_over_issued_iscore, icx_issue_amount

    @classmethod
    def _correct_issue_amount(cls,
                              remain_over_issued_iscore: int,
                              icx_issue_amount: int,
                              prev_block_cumulative_fee: int) -> Tuple[int, int, int, int]:
        assert icx_issue_amount >= 0
        assert prev_block_cumulative_fee >= 0

        over_issued_icx, over_issued_iscore = cls._separate_icx_and_iscore(remain_over_issued_iscore)

        # covered_icx_by_remain is always positive value
        covered_icx_by_fee, covered_icx_by_remain, remain_over_issued_icx, icx_issue_amount = \
            cls._reflect_difference_in_issuing(icx_issue_amount, over_issued_icx, prev_block_cumulative_fee)
        remain_over_issued_iscore: int = remain_over_issued_icx * ISCORE_EXCHANGE_RATE + over_issued_iscore

        return covered_icx_by_fee, covered_icx_by_remain, remain_over_issued_iscore, icx_issue_amount



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from shutil import copytree
from typing import TYPE_CHECKING

from iconcommons.logger import Logger
from iconservice.base.address import Address
from iconservice.base.exception import AccessDeniedException
from iconservice.deploy.storage import IconScoreDeployInfo, DeployState
from iconservice.deploy.utils import remove_path
from iconservice.icon_constant import BUILTIN_SCORE_ADDRESS_MAPPER, ZERO_TX_HASH, ICON_DEPLOY_LOG_TAG
from iconservice.iconscore.icon_score_context_util import IconScoreContextUtil

if TYPE_CHECKING:
    from iconservice.iconscore.icon_score_base import IconScoreBase
    from iconservice.iconscore.icon_score_context import IconScoreContext
    from iconservice.iconscore.icon_score_mapper_object import IconScoreInfo


class IconBuiltinScoreLoader(object):
    """Before activating icon_service_engine, deploy builtin scores which has never been deployed.
    """

    @staticmethod
    def _pre_builtin_score_root_path():
        root_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))
        return os.path.join(root_path, 'builtin_scores')

    @staticmethod
    def load_builtin_scores(context: 'IconScoreContext', builtin_score_owner: 'Address'):
        for score_name, value in BUILTIN_SCORE_ADDRESS_MAPPER.items():
            score_address = Address.from_string(value)

            # If builtin score has been already deployed, exit.
            if IconScoreContextUtil.is_score_active(context, score_address):
                IconBuiltinScoreLoader._load_score(context, score_address, builtin_score_owner)
            else:
                IconBuiltinScoreLoader._deploy_score(
                    context, score_name, score_address, builtin_score_owner)

    @staticmethod
    def _load_score(context: 'IconScoreContext',
                    score_address: 'Address', builtin_score_owner: 'Address'):
        score: 'IconScoreBase' = IconScoreContextUtil.get_icon_score(context, score_address)
        assert score is not None

        if score.owner != builtin_score_owner:
            raise AccessDeniedException(
                f'score.owner({score.owner}) != builtin_score_owner({builtin_score_owner})')

    @staticmethod
    def _deploy_score(context: 'IconScoreContext',
                      score_name: str,
                      score_address: 'Address',
                      builtin_score_owner: 'Address'):

        score_source_path_in_iconservice: str = os.path.join(
            IconBuiltinScoreLoader._pre_builtin_score_root_path(), score_name)

        # Save deploy_info for a builtin score to storage.
        deploy_info = IconScoreDeployInfo(
            score_address=score_address,
            deploy_state=DeployState.ACTIVE,
            owner=builtin_score_owner,
            current_tx_hash=ZERO_TX_HASH,
            next_tx_hash=ZERO_TX_HASH)
        context.storage.deploy.put_deploy_info(context, deploy_info)

        tx_hash: bytes = deploy_info.current_tx_hash

        # score_path is score_root_path/score_address/ directory.
        score_path: str = os.path.join(
            context.score_root_path, score_address.to_bytes().hex())

        # Make a directory for a builtin score with a given score_address.
        os.makedirs(score_path, exist_ok=True)

        try:
            score_deploy_path: str = os.path.join(score_path, f'0x{tx_hash.hex()}')

            # remove_path() supports directory as well as file.
            remove_path(score_deploy_path)
            # Copy builtin score source files from iconservice package to score_deploy_path.
            copytree(score_source_path_in_iconservice, score_deploy_path)
        except FileExistsError:
            pass

        try:
            # Import score class from deployed builtin score sources
            score_info: 'IconScoreInfo' =\
                IconScoreContextUtil.create_score_info(context, score_address, tx_hash)

            # Create a score instance from the imported score class.
            score = score_info.create_score()

            # Call on_install() to initialize the score database of the builtin score.
            score.on_install()
        except BaseException as e:
            Logger.exception(
                f'Failed to deploy a builtin score: {score_address}\n{str(e)}',
                ICON_DEPLOY_LOG_TAG)
            raise e



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
from typing import TYPE_CHECKING, Optional, List, Dict, Iterable

from iconcommons import Logger

from .base.block import Block, NULL_BLOCK
from .base.exception import InvalidParamsException, InternalServiceErrorException
from .database.batch import BlockBatch
from .icon_constant import Revision
from .iconscore.icon_score_mapper import IconScoreMapper
from .iiss.reward_calc.msg_data import TxData
from .inv.container import Container as INVContainer
from .prep.prep_address_converter import PRepAddressConverter
from .utils import bytes_to_hex, sha3_256, to_camel_case
from . import __version__

if TYPE_CHECKING:
    from .base.address import Address
    from .prep.data import PRepContainer, Term

_TAG = "PRECOMMIT"


class PrecommitDataWriter:
    """
    Write Precommit data to the file for debugging
    """
    VERSION = 0
    DIR_NAME = "precommit"

    def __init__(self, path: str):
        self._dir_path: str = os.path.join(path, self.DIR_NAME)
        if not os.path.exists(self._dir_path):
            os.makedirs(self._dir_path)
        self._filename_suffix = f"precommit-v{self.VERSION}.json"

    def write(self, precommit_data: 'PrecommitData'):
        """
            Write the human readable precommit data to the file for debugging
            :param precommit_data:
            :param path: path to record
            :return:
            """
        filename: str = f"{precommit_data.block.height}" \
                        f"-{precommit_data.state_root_hash.hex()[:8]}" \
                        f"-{self._filename_suffix}"
        Logger.info(
            tag=_TAG,
            msg=f"PrecommitDataWriter.write() start (precommit: {precommit_data})"
        )
        with open(os.path.join(self._dir_path, filename), 'w') as f:
            try:
                block = precommit_data.block
                json_dict = {
                    "iconservice": __version__,
                    "revision": precommit_data.revision,
                    "block": block.to_dict(to_camel_case) if block is not None else None,
                    "isStateRootHash": precommit_data.is_state_root_hash,
                    "rcStateRootHash": precommit_data.rc_state_root_hash,
                    "stateRootHash": precommit_data.state_root_hash,
                    "prevBlockGenerator": precommit_data.prev_block_generator,
                    "blockBatch": precommit_data.block_batch.to_list(),
                    "rcBlockBatch": self._convert_rc_block_batch_to_list(precommit_data.rc_block_batch)
                }
                json.dump(json_dict, f, default=self._json_default)
            except Exception as e:
                Logger.exception(
                    tag=_TAG,
                    msg=f"Exception raised during writing the precommit-data: {e}"
                )
        Logger.info(
            tag=_TAG,
            msg=f"PrecommitDataWriter.write() end"
        )

    @classmethod
    def _json_default(cls, obj):
        # json package do not  Address obj
        from iconservice import Address

        if isinstance(obj, bytes):
            return bytes_to_hex(obj)
        elif isinstance(obj, Address):
            return str(obj)
        return obj

    @classmethod
    def _convert_rc_block_batch_to_list(cls, rc_block_batch: list):
        new_list = []
        tx_index = 0
        for data in rc_block_batch:
            if isinstance(data, TxData):
                key: bytes = data.make_key(tx_index)
                tx_index += 1
            else:
                key: bytes = data.make_key()

            value: bytes = data.make_value()
            new_list.append({
                "key": key,
                "value": value
            })
        return new_list


class PrecommitData(object):
    def __init__(self,
                 revision: int,
                 rc_db_revision: int,
                 inv_container: 'INVContainer',
                 block_batch: 'BlockBatch',
                 block_result: list,
                 rc_block_batch: list,
                 preps: 'PRepContainer',
                 term: Optional['Term'],
                 prev_block_generator: Optional['Address'],
                 prev_block_validators: Optional[List['Address']],
                 score_mapper: Optional['IconScoreMapper'],
                 rc_state_root_hash: Optional[bytes],
                 added_transactions: dict,
                 next_preps: Optional[dict],
                 prep_address_converter: 'PRepAddressConverter'):
        """

        :param block_batch: changed states for a block
        :param block_result: tx_results made from transactions in a block
        :param score_mapper: newly deployed scores in a block

        """
        # Todo: check if remove the revision
        self.revision: int = revision
        self.rc_db_revision: int = rc_db_revision
        self.inv_container: 'INVContainer' = inv_container
        self.block_batch = block_batch
        self.block_result = block_result
        self.rc_block_batch = rc_block_batch
        # Snapshot of preps
        self.preps = preps
        self.term = term
        self.prev_block_generator = prev_block_generator
        self.prev_block_validators = prev_block_validators
        self.score_mapper = score_mapper

        self.is_state_root_hash: bytes = self.block_batch.digest()
        self.rc_state_root_hash: Optional[bytes] = rc_state_root_hash

        self.state_root_hash: bytes = self._make_state_root_hash()

        self.added_transactions: dict = added_transactions
        self.next_preps: Optional[dict] = next_preps

        self.prep_address_converter: 'PRepAddressConverter' = prep_address_converter

        # To prevent redundant precommit data logging
        self.already_exists = False

        # Make preps and term immutable
        if preps:
            preps.freeze()
        if term:
            term.freeze()

    def __str__(self):
        lines = [
            f"revision: {self.revision}",
            f"block: {self.block}",
            f"is_state_root_hash: {bytes_to_hex(self.is_state_root_hash)}",
            f"rc_state_root_hash: {bytes_to_hex(self.rc_state_root_hash)}",
            f"state_root_hash: {bytes_to_hex(self.state_root_hash)}",
            f"prev_block_generator: {self.prev_block_generator}",
            f"added_transactions: {self.added_transactions}",
            f"next_preps: {self.next_preps}"
        ]
        return "\n".join(lines)

    @property
    def block(self) -> Optional['Block']:
        return None if self.block_batch is None else self.block_batch.block

    def _make_state_root_hash(self) -> bytes:
        if self.revision < Revision.DECENTRALIZATION.value or self.rc_state_root_hash is None:
            return self.is_state_root_hash

        data = [self.is_state_root_hash, self.rc_state_root_hash]
        value: bytes = b'|'.join(data)
        return sha3_256(value)


class PrecommitDataManager(object):
    """Manages multiple precommit block data

    Assume that it manages only 2-depth block precommit
    - root: confirmed block: 0-depth
    - root.children (parent): 1-depth
    - parent.children: 2-depth
    """

    class Node(object):
        """Inner class used to manage PrecommitData in LinkedList

        """

        def __init__(
                self, block: 'Block',
                data: 'PrecommitData' = None,
                parent: Optional['PrecommitDataManager.Node'] = None):
            self._block = block
            self._data = data
            self._parent = parent
            self._children: Dict[bytes, 'PrecommitDataManager.Node'] = {}

        @property
        def precommit_data(self) -> 'PrecommitData':
            return self._data

        @property
        def parent(self) -> Optional['PrecommitDataManager.Node']:
            return self._parent

        @parent.setter
        def parent(self, node: Optional['PrecommitDataManager.Node']):
            self._parent = node

        @property
        def block(self) -> 'Block':
            return self._block

        def add_child(self, node: 'PrecommitDataManager.Node'):
            self._children[node.block.hash] = node

        def get_child(self, block_hash: bytes) -> Optional['PrecommitDataManager.Node']:
            return self._children.get(block_hash)

        def children(self) -> Iterable['PrecommitDataManager.Node']:
            for child in self._children.values():
                yield child

        def is_root(self) -> bool:
            return self._parent is None

        def is_leaf(self) -> bool:
            return len(self._children) == 0

        def update_block_hash(self, block_hash: bytes):
            self.precommit_data.block_batch.update_block_hash(block_hash)
            self._block = self.precommit_data.block_batch.block

    def __init__(self):
        self._root: Optional['PrecommitDataManager.Node'] = None
        # block_hash : PrecommitDataManager.Node instance
        self._precommit_data_mapper: Dict[bytes, 'PrecommitDataManager.Node'] = {}

    def __len__(self):
        return len(self._precommit_data_mapper)

    def init(self, block: Optional['Block']):
        """This is called in IconServiceEngine.open(), IconServiceEngine.rollback() or self._clear()

        :param block: the last confirmed(=committed) block
        :return:
        """
        if not block:
            block = NULL_BLOCK

        self._precommit_data_mapper.clear()

        root = PrecommitDataManager.Node(block)
        self._precommit_data_mapper[root.block.hash] = root
        self._set_root(root)

    @property
    def last_block(self) -> 'Block':
        return self._root.block

    def push(self, precommit_data: 'PrecommitData'):
        block: 'Block' = precommit_data.block_batch.block
        parent: Optional['PrecommitDataManager.Node'] = self._precommit_data_mapper.get(block.prev_hash)

        if parent is None:
            # Genesis block has no prev(=parent) block
            parent = self._root
        elif not (parent and parent.block.height == block.height - 1):
            raise InternalServiceErrorException(
                f"Parent precommitData not found: {bytes_to_hex(block.prev_hash)}")

        node = PrecommitDataManager.Node(block, precommit_data, parent)

        parent.add_child(node)
        self._precommit_data_mapper[block.hash] = node

    def get(self, block_hash: bytes) -> Optional['PrecommitData']:
        if not isinstance(block_hash, bytes):
            return None

        node = self._precommit_data_mapper.get(block_hash)
        return node.precommit_data if node else None

    def commit(self, block: 'Block'):
        node = self._precommit_data_mapper.get(block.hash)
        if node is None:
            Logger.warning(
                tag=_TAG,
                msg=f"No precommit data: height={block.height} hash={bytes_to_hex(block.hash)}")
            return

        if not node.parent.is_root() and self._root == node.parent:
            raise InternalServiceErrorException(f"Parent should be a root")

        self._remove_sibling_precommit_data(block)
        del self._precommit_data_mapper[self._root.block.hash]
        self._set_root(node)

    def _remove_sibling_precommit_data(self, block_to_commit: 'Block'):
        """Remove the sibling blocks whose height is the same as that of the block to commit

        :param block_to_commit:
        :return:
        """

        def pick_up_blocks_to_remove() -> Iterable['PrecommitDataManager.Node']:
            for parent in self._root.children():
                if block_to_commit.hash == parent.block.hash:
                    # DO NOT remove the confirmed block and its children
                    continue

                yield parent
                for child in parent.children():
                    assert parent == child.parent
                    assert child.is_leaf()
                    yield child

        for node_to_remove in pick_up_blocks_to_remove():
            del self._precommit_data_mapper[node_to_remove.block.hash]

    # def remove_precommit_state(self, instant_block_hash: bytes):
    #     if instant_block_hash in self._precommit_data_mapper:
    #         del self._precommit_data_mapper[instant_block_hash]

    def clear(self):
        """Clear all data except for root

        :return:
        """
        self.init(self.last_block)

    def validate_block_to_invoke(self, block: 'Block'):
        """Check if the block to invoke is valid before invoking it

        :param block: block to invoke
        """
        if self._root.block.height < 0:
            # Exception handling for genesis block
            return

        parent: 'PrecommitDataManager.Node' = self._precommit_data_mapper.get(block.prev_hash)
        if parent:
            if block.prev_hash == parent.block.hash and block.height == parent.block.height + 1:
                return

        raise InvalidParamsException(
            f'Failed to invoke a block: '
            f'prev_block({parent.block if parent else None}) '
            f'block_to_invoke({block})')

    def validate_block_to_commit(self, block_hash: bytes):
        """Check block validation
        before write_precommit_state() or remove_precommit_state()
        No need to consider an instant_block_hash on 2-depth block invocation

        :param block_hash: hash data which is used for retrieving block instance from the pre-commit data mapper
        """
        assert isinstance(block_hash, bytes)

        node: 'PrecommitDataManager.Node' = self._precommit_data_mapper.get(block_hash)
        if node is None:
            raise InvalidParamsException(
                f'No precommit data: block_hash={bytes_to_hex(block_hash)}')

        block = node.block
        prev_block = self._root.block

        if block.height == prev_block.height + 1 \
                and (block.height == 0 or node.block.prev_hash == prev_block.hash):
            return

        raise InvalidParamsException(
            f'Invalid precommit block: prev_block({prev_block}) block({block})')

    def _set_root(self, node: 'PrecommitDataManager.Node'):
        node.parent = None
        self._root = node

    def get_block_batches(self, block_hash: bytes) -> Iterable['BlockBatch']:
        node = self._precommit_data_mapper.get(block_hash)
        if not node:
            return

        while not node.is_root():
            yield node.precommit_data.block_batch
            # parent means previous block node
            node = node.parent

    def change_block_hash(self, block_height: int, src: bytes, dst: bytes):
        node: 'PrecommitDataManager.Node' = self._precommit_data_mapper.get(src)
        if node is None:
            raise InvalidParamsException(
                f'No precommit data: block_hash={bytes_to_hex(src)}')

        if not node.is_leaf() or node.block.height != block_height:
            raise InvalidParamsException(
                f"Invalid node data (not leaf): src={src}, dest={dst}")

        node.update_block_hash(block_hash=dst)
        del self._precommit_data_mapper[src]
        self._precommit_data_mapper[dst] = node



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import asyncio
import copy
import os
import signal
import sys

import aio_pika
import pkg_resources
import setproctitle
from earlgrey import MessageQueueService

from iconcommons.icon_config import IconConfig
from iconcommons.logger import Logger
from iconservice.base.exception import FatalException
from iconservice.icon_config import default_icon_config, check_config, args_to_dict
from iconservice.icon_constant import ICON_SERVICE_PROCTITLE_FORMAT, ICON_SCORE_QUEUE_NAME_FORMAT, ConfigKey
from iconservice.icon_inner_service import IconScoreInnerService
from iconservice.icon_service_cli import ExitCode

_TAG = 'CLI'


class IconService(object):

    """IconScore service for stand alone start.
    It provides gRPC interface for peer_service to communicate with icon service.
    Its role is the bridge between loopchain and IconServiceEngine.
    """

    def __init__(self):
        self._icon_score_queue_name = None
        self._amqp_target = None
        self._inner_service = None

    def serve(self, config: 'IconConfig'):
        async def _serve():
            await self._inner_service.connect(exclusive=True)
            Logger.info(f'Start IconService Service serve!', _TAG)

        channel = config[ConfigKey.CHANNEL]
        amqp_key = config[ConfigKey.AMQP_KEY]
        amqp_target = config[ConfigKey.AMQP_TARGET]
        score_root_path = config[ConfigKey.SCORE_ROOT_PATH]
        db_root_path = config[ConfigKey.STATE_DB_ROOT_PATH]
        version: str = get_version()

        self._set_icon_score_stub_params(channel, amqp_key, amqp_target)

        Logger.info(f'==========IconService Service params==========', _TAG)

        Logger.info(f'version : {version}', _TAG)
        Logger.info(f'score_root_path : {score_root_path}', _TAG)
        Logger.info(f'icon_score_state_db_root_path  : {db_root_path}', _TAG)
        Logger.info(f'amqp_target  : {amqp_target}', _TAG)
        Logger.info(f'amqp_key  :  {amqp_key}', _TAG)
        Logger.info(f'icon_score_queue_name  : {self._icon_score_queue_name}', _TAG)
        Logger.info(f'==========IconService Service params==========', _TAG)

        # Before creating IconScoreInnerService instance,
        # loop SHOULD be set as a current event loop for the current thread.
        # Otherwise connection between iconservice and rc will be failed.
        loop = MessageQueueService.loop
        asyncio.set_event_loop(loop)

        try:
            self._inner_service = IconScoreInnerService(amqp_target, self._icon_score_queue_name, conf=config)
        except FatalException as e:
            Logger.exception(f"{e}", _TAG)
            Logger.error(f"{e}", _TAG)
            self._inner_service.clean_close()

        loop.create_task(_serve())
        loop.add_signal_handler(signal.SIGINT, self.signal_handler, signal.SIGINT)
        loop.add_signal_handler(signal.SIGTERM, self.signal_handler, signal.SIGTERM)

        try:
            loop.run_forever()
        except FatalException as e:
            Logger.exception(f"{e}", _TAG)
            Logger.error(f"{e}", _TAG)
            self._inner_service.clean_close()
        finally:
            """
            If the function is called when the operation is not an endless loop 
            in an asynchronous function, the await is terminated immediately.
            """
            Logger.info(f"loop has been stopped and will be closed.")

            loop.run_until_complete(loop.shutdown_asyncgens())

            self.cancel_tasks(loop)

            # close icon service components
            self._inner_service.clean_close()

            loop.close()

    @staticmethod
    def cancel_tasks(loop):
        pending = asyncio.Task.all_tasks(loop)
        Logger.info(f"cancel pending {len(pending)} tasks in event loop.")
        for task in pending:
            if task.done():
                continue
            task.cancel()
            try:
                loop.run_until_complete(task)
            except asyncio.CancelledError as e:
                Logger.info(f"cancel pending task: {task}, error: {e}")

    @staticmethod
    def signal_handler(signum: int):
        Logger.debug(f"Get signal {signum}")
        asyncio.get_event_loop().stop()

    def _set_icon_score_stub_params(self, channel: str, amqp_key: str, amqp_target: str):
        self._icon_score_queue_name = \
            ICON_SCORE_QUEUE_NAME_FORMAT.format(channel_name=channel, amqp_key=amqp_key)
        self._amqp_target = amqp_target


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-sc", dest=ConfigKey.SCORE_ROOT_PATH, type=str, default=None,
                        help="icon score root path  example : .score")
    parser.add_argument("-st", dest=ConfigKey.STATE_DB_ROOT_PATH, type=str, default=None,
                        help="icon score state db root path  example : .statedb")
    parser.add_argument("-ch", dest=ConfigKey.CHANNEL, type=str, default=None,
                        help="icon score channel")
    parser.add_argument("-ak", dest=ConfigKey.AMQP_KEY, type=str, default=None,
                        help="icon score amqp_key : [amqp_key]")
    parser.add_argument("-at", dest=ConfigKey.AMQP_TARGET, type=str, default=None,
                        help="icon score amqp_target : [127.0.0.1]")
    parser.add_argument("-c", dest=ConfigKey.CONFIG, type=str, default=None,
                        help="icon score config")
    parser.add_argument("-tbears", dest=ConfigKey.TBEARS_MODE, action='store_true',
                        help="tbears mode")
    parser.add_argument("-steptrace", dest=ConfigKey.STEP_TRACE_FLAG, action="store_true", help="enable step tracing")

    args = parser.parse_args()

    args_params = dict(vars(args))
    del args_params['config']
    setproctitle.setproctitle(ICON_SERVICE_PROCTITLE_FORMAT.format(**args_params))

    conf_path = args.config
    if conf_path is not None:
        if not IconConfig.valid_conf_path(conf_path):
            print(f'invalid config file : {conf_path}')
            sys.exit(ExitCode.INVALID_COMMAND.value)
    if conf_path is None:
        conf_path = str()

    conf = IconConfig(conf_path, copy.deepcopy(default_icon_config))
    conf.load()
    conf.update_conf(args_to_dict(args))
    Logger.load_config(conf)
    if not check_config(conf, default_icon_config):
        Logger.error(tag=_TAG, msg=f"Invalid Config")
        sys.exit(ExitCode.INVALID_CONFIG.value)

    Logger.print_config(conf, _TAG)

    _run_async(_check_rabbitmq(conf[ConfigKey.AMQP_TARGET]))
    icon_service = IconService()
    icon_service.serve(config=conf)
    Logger.info(f'==========IconService Done==========', _TAG)


def run_in_foreground(conf: 'IconConfig'):
    _run_async(_check_rabbitmq(conf[ConfigKey.AMQP_TARGET]))
    icon_service = IconService()
    icon_service.serve(config=conf)


def _run_async(async_func):
    loop = MessageQueueService.loop
    return loop.run_until_complete(async_func)


async def _check_rabbitmq(amqp_target: str):
    connection = None
    try:
        amqp_user_name = os.getenv("AMQP_USERNAME", "guest")
        amqp_password = os.getenv("AMQP_PASSWORD", "guest")
        connection = await aio_pika.connect(host=amqp_target, login=amqp_user_name, password=amqp_password)
    except ConnectionRefusedError:
        Logger.error("rabbitmq-service disable", _TAG)
        exit(0)
    finally:
        if connection:
            await connection.close()


DIR_PATH = os.path.abspath(os.path.dirname(__file__))
PROJECT_ROOT_PATH = os.path.abspath(os.path.join(DIR_PATH, '..'))


def get_version() -> str:
    """Get version of iconservice.
    The location of the file that holds the version information is different when packaging and when executing.
    :return: version of tbears.
    """
    try:
        version = pkg_resources.get_distribution('iconservice').version
    except pkg_resources.DistributionNotFound:
        version_path = os.path.join(PROJECT_ROOT_PATH, 'VERSION')
        with open(version_path, mode='r') as version_file:
            version = version_file.read()
    except:
        version = 'unknown'
    return version


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""IconScoreEngine module
"""

from copy import deepcopy
from typing import TYPE_CHECKING, Any

from .icon_score_constant import STR_FALLBACK, ATTR_SCORE_GET_API, ATTR_SCORE_CALL
from .icon_score_context import IconScoreContext
from .icon_score_context_util import IconScoreContextUtil
from .typing.conversion import convert_score_parameters, ConvertOption
from .typing.element import (
    ScoreElementMetadata,
    get_score_element_metadata,
)
from ..base.address import Address, SYSTEM_SCORE_ADDRESS
from ..base.exception import ScoreNotFoundException, InvalidParamsException
from ..icon_constant import Revision, DataType

if TYPE_CHECKING:
    from ..iconscore.icon_score_base import IconScoreBase


class IconScoreEngine(object):
    """Calls external functions provided by each IconScore
    """

    @staticmethod
    def invoke(context: 'IconScoreContext',
               icon_score_address: 'Address',
               data_type: str,
               data: dict) -> None:
        """Handle calldata contained in icx_sendTransaction message

        :param icon_score_address:
        :param context:
        :param data_type:
        :param data: calldata
        """
        IconScoreEngine._validate_score_blacklist(context, icon_score_address)

        if data_type == DataType.CALL:
            IconScoreEngine._call(context, icon_score_address, data)
        else:
            IconScoreEngine._fallback(context, icon_score_address)

    @staticmethod
    def query(context: IconScoreContext,
              icon_score_address: Address,
              data_type: str,
              data: dict) -> object:
        """Execute an external method of SCORE without state changing

        Handles messagecall of icx_call
        """
        IconScoreEngine._validate_score_blacklist(context, icon_score_address)

        if data_type == DataType.CALL:
            return IconScoreEngine._call(context, icon_score_address, data)
        else:
            raise InvalidParamsException(f'Invalid dataType: {data_type}')

    @staticmethod
    def get_score_api(context: 'IconScoreContext', icon_score_address: 'Address') -> object:
        """Handle get score api

        :param context:
        :param icon_score_address:
        """
        IconScoreEngine._validate_score_blacklist(context, icon_score_address)

        icon_score = IconScoreEngine._get_icon_score(context, icon_score_address)
        get_api = getattr(icon_score, ATTR_SCORE_GET_API)
        return get_api()

    @staticmethod
    def _validate_score_blacklist(context: 'IconScoreContext', icon_score_address: 'Address'):
        if icon_score_address is None or not icon_score_address.is_contract:
            raise InvalidParamsException(f"Invalid score address: ({icon_score_address})")

        IconScoreContextUtil.validate_score_blacklist(context, icon_score_address)

    @classmethod
    def _call(cls, context: 'IconScoreContext',
              icon_score_address: 'Address',
              data: dict) -> Any:
        """Handle jsonrpc including both invoke and query

        :param context:
        :param icon_score_address:
        :param data: data to call the method of score
        """
        func_name: str = data['method']
        kw_params: dict = data.get('params', {})

        icon_score = cls._get_icon_score(context, icon_score_address)

        converted_params = cls._convert_score_params_by_annotations(
            context, icon_score, func_name, kw_params)
        context.set_func_type_by_icon_score(icon_score, func_name)
        context.current_address = icon_score_address

        score_func = getattr(icon_score, ATTR_SCORE_CALL)
        ret = score_func(func_name=func_name, kw_params=converted_params)

        # No problem even though ret is None
        return deepcopy(ret)

    @classmethod
    def _convert_score_params_by_annotations(
            cls, context: 'IconScoreContext',
            icon_score: 'IconScoreBase',
            func_name: str,
            kw_params: dict) -> dict:

        options = ConvertOption.NONE
        if (
                icon_score.address == SYSTEM_SCORE_ADDRESS
                and func_name in ("registerPRep", "setPRep")
                and context.revision < Revision.SCORE_FUNC_PARAMS_CHECK.value
        ):
            options = ConvertOption.IGNORE_UNKNOWN_PARAMS

        element_metadata: ScoreElementMetadata = get_score_element_metadata(icon_score, func_name)
        params = convert_score_parameters(kw_params, element_metadata.signature, options)

        return params

    @staticmethod
    def _fallback(context: 'IconScoreContext',
                  score_address: 'Address'):
        """When an IconScore receives some coins and calldata is None,
        fallback function is called.

        :param score_address:
        """
        icon_score = IconScoreEngine._get_icon_score(context, score_address)

        score_func = getattr(icon_score, ATTR_SCORE_CALL)
        score_func(STR_FALLBACK)

    @staticmethod
    def _get_icon_score(context: 'IconScoreContext', icon_score_address: 'Address'):
        icon_score = IconScoreContextUtil.get_icon_score(context, icon_score_address)
        if icon_score is None:
            raise ScoreNotFoundException(
                f'SCORE not found: {icon_score_address}')
        return icon_score



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from collections import namedtuple
from typing import Any, Dict, List

from .data.value import Value, VALUE_MAPPER
from .. import Address
from ..base.exception import InvalidParamsException
from ..icon_constant import IconNetworkValueType, IconScoreContextType
from ..iconscore.icon_score_step import StepType

SystemRevision = namedtuple('SystemRevision', ['code', 'name'])


class ValueConverter(object):
    @staticmethod
    def convert_for_icon_service(type_: 'IconNetworkValueType', value: Any) -> 'Value':
        """
        Convert IconNetwork value data type for icon service.
        Some data need to be converted for enhancing efficiency.
        :param type_:
        :param value:
        :return:
        """
        converted_value: Any = value
        if type_ == IconNetworkValueType.MAX_STEP_LIMITS:
            converted_value: dict = {}
            for key, value in value.items():
                if isinstance(key, str):
                    if key == "invoke":
                        converted_value[IconScoreContextType.INVOKE] = value
                    elif key == "query":
                        converted_value[IconScoreContextType.QUERY] = value
                    else:
                        raise InvalidParamsException(f"Invalid context type: {type_.name}")
                else:
                    raise ValueError(f"Invalid data type: "
                                     f"value: {type_.name} "
                                     f"key type: {type(key)}")
        elif type_ == IconNetworkValueType.STEP_COSTS:
            converted_value: dict = {}
            for key, value in value.items():
                if isinstance(key, str):
                    try:
                        converted_value[StepType(key)] = value
                    except ValueError:
                        # Pass the unknown step type
                        pass
                else:
                    raise ValueError(f"Invalid data type: "
                                     f"value: {type_.name} "
                                     f"key type: {type(key)}")
        try:
            value: 'Value' = VALUE_MAPPER[type_](converted_value)
        except KeyError:
            raise InvalidParamsException(f"Invalid INV key: {type_}")

        return value

    @staticmethod
    def convert_for_governance(type_: 'IconNetworkValueType', value: Any) -> Any:
        """
        Convert IconNetwork value data type for governance score
        Some data which have been converted for enhancing efficiency need to be converted.

        :param type_:
        :param value:
        :return:
        """
        converted_value: Any = value
        if type_ == IconNetworkValueType.MAX_STEP_LIMITS:
            converted_value: dict = {}
            for key, value in value.items():
                assert isinstance(key, IconScoreContextType)
                converted_value[key.name.lower()] = value
        elif type_ == IconNetworkValueType.STEP_COSTS:
            converted_value: dict = {}
            for key, value in value.items():
                assert isinstance(key, StepType)
                converted_value[key.value] = value
        return converted_value


class Container(object):
    class BatchDict(dict):
        def __init__(self):
            super().__init__()
            self._migration_trigger: bool = False

        def __setitem__(self, key, value):
            if value is None:
                return
            if not isinstance(key, IconNetworkValueType) or key not in IconNetworkValueType:
                raise ValueError(f"Invalid value key: {key}")
            if not isinstance(value, Value):
                raise ValueError(f"Invalid value type: {type(value)}")

            super().__setitem__(key, value)

        def trigger_migration(self):
            self._migration_trigger: bool = True

        def is_migration_triggered(self):
            return self._migration_trigger

    def __init__(self, is_migrated: bool):
        # Todo: Freeze data
        self._is_migrated: bool = is_migrated

        self._tx_batch = self.BatchDict()
        self._icon_network_values: dict = {}    # It must be set by DB loading and network proposal

    @property
    def is_migrated(self) -> bool:
        return self._is_migrated

    def get_by_type(self, type_: 'IconNetworkValueType') -> Any:
        try:
            value = self._tx_batch.get(type_, self._icon_network_values[type_])
            if isinstance(value, Value):
                return value.value
        except KeyError:
            return None
        return None

    @property
    def service_config(self) -> int:
        return self.get_by_type(IconNetworkValueType.SERVICE_CONFIG)

    @property
    def step_price(self) -> int:
        return self.get_by_type(IconNetworkValueType.STEP_PRICE)

    @property
    def step_costs(self) -> Dict['StepType', int]:
        return self.get_by_type(IconNetworkValueType.STEP_COSTS)

    @property
    def max_step_limits(self) -> Dict['IconScoreContextType', int]:
        return self.get_by_type(IconNetworkValueType.MAX_STEP_LIMITS)

    @property
    def revision_code(self) -> int:
        return self.get_by_type(IconNetworkValueType.REVISION_CODE)

    @property
    def revision_name(self) -> str:
        return self.get_by_type(IconNetworkValueType.REVISION_NAME)

    @property
    def score_black_list(self) -> List['Address']:
        return self.get_by_type(IconNetworkValueType.SCORE_BLACK_LIST)

    @property
    def import_white_list(self) -> Dict[str, List[str]]:
        return self.get_by_type(IconNetworkValueType.IMPORT_WHITE_LIST)

    @property
    def irep(self) -> int:
        return self.get_by_type(IconNetworkValueType.IREP)

    def update_batch(self):
        for value in self._tx_batch.values():
            self._set(value)

    def clear_batch(self):
        self._tx_batch.clear()

    def migrate(self, data: List['Value']):
        """
        Migrates governance variable from SCORE DB to State DB.
        It will be called when updating governance score to version "".
        This method is called only once.
        :param data:
        :return:
        """
        if len(data) != IconNetworkValueType.gs_migration_count():
            raise InvalidParamsException("Migration data for Icon Network Values are insufficient")

        for value in data:
            self._tx_batch[value.TYPE] = value
        self._tx_batch.trigger_migration()

    def update_migration_if_succeed(self):
        if self._tx_batch.is_migration_triggered():
            self._is_migrated = True
            self.update_batch()

    def _set(self, value: 'Value'):
        self._icon_network_values[value.TYPE] = value

    def set_inv(self, value: 'Value', is_open: bool = False):
        """
        Set value on ICON Network Value from icon service.
        There are two cases of calling this method.
        First: Before migration
        Second: Initiating 'system value' when opening icon service (i.e. first initiation)
        Third:

        :param value:
        :param is_open:
        :return:
        """
        if not self._is_migrated or is_open is True:
            self._set(value)
        else:
            raise PermissionError(f"Invalid case of setting ICON Network value from icon-service"
                                  f"migration: {self._is_migrated} is open: {is_open}")

    def set_inv_to_tx_batch(self, value: 'Value'):
        """
        Set values on ICON Network Value and put these into DB.
        Only Governance SCORE can set values after migration.
        :param value:
        :return:
        """
        assert self._is_migrated
        # Update member variables
        # Check If value is valid
        self._tx_batch[value.TYPE] = value

    def copy(self) -> 'Container':
        """Copy container"""
        container = copy.copy(self)
        container._tx_batch = self.BatchDict()
        container._icon_network_values = copy.copy(self._icon_network_values)
        return container



# -*- coding: utf-8 -*-
# Copyright 2020 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = (
    "is_base_type",
    "name_to_type",
    "get_origin",
    "get_args",
    "is_struct",
    "get_annotations",
)

from typing import Tuple, Union, Type, Dict, Any, Optional

from iconservice.base.address import Address

BaseObject = Union[bool, bytes, int, str, 'Address']
BaseObjectType = Type[BaseObject]

BASE_TYPES = {bool, bytes, int, str, Address}
TYPE_NAME_TO_TYPE = {_type.__name__: _type for _type in BASE_TYPES}


def is_base_type(value: type) -> bool:
    try:
        return value in BASE_TYPES
    except:
        return False


def name_to_type(type_name: str) -> BaseObjectType:
    return TYPE_NAME_TO_TYPE[type_name]


def get_origin(type_hint: type) -> Optional[type]:
    """
    Dict[str, int] -> dict
    List[str] -> list
    subclass of type -> itself
    subclass of TypedDict -> itself

    :param type_hint:
    :return:
    """
    if type_hint == "Address":
        type_hint = Address

    if isinstance(type_hint, type):
        return type_hint

    return getattr(type_hint, "__origin__", None)


def get_args(type_hint: type) -> Tuple[type]:
    return getattr(type_hint, "__args__", ())


def is_struct(type_hint) -> bool:
    try:
        return type_hint.__class__.__name__ == "_TypedDictMeta"
    except:
        return False


def get_annotations(obj: Any, default: Any) -> Dict[str, type]:
    return getattr(obj, "__annotations__", default)


def isinstance_ex(value: Any, _type: type) -> bool:
    if not isinstance(value, _type):
        return False

    if type(value) is bool and _type is not bool:
        return False

    return True



# -*- coding: utf-8 -*-

from typing import Optional

from ..base.address import Address
from ..base.exception import InvalidBalanceException
from ..utils import bytes_to_hex


class BalanceVerifier(object):
    """Check for balance integrity
    """

    def __init__(self):
        self._tx_hash: Optional[bytes] = None
        self._issue = 0
        self._balance = 0

    def open(self, tx_hash: bytes):
        self._tx_hash = tx_hash
        self.reset()

    def issue(self, value: int):
        assert value >= 0
        self._issue += value

    def burn(self, value: int):
        assert value >= 0
        self._issue -= value

    def deposit(self, _address: Address, value: int):
        assert value >= 0
        self._balance += value

    def withdraw(self, _address: Address, value: int):
        assert value >= 0
        self._balance -= value

    def reset(self):
        """Reset the values of issue and balance to 0

        This is called when invoking a tx gets started or a tx is failed
        """
        self._issue = 0
        self._balance = 0

    def verify(self):
        if self._balance != self._issue:
            raise InvalidBalanceException(
                f"Invalid asset integrity: "
                f"tx_hash={bytes_to_hex(self._tx_hash)} "
                f"issue={self._issue} "
                f"balance={self._balance}"
            )

    def close(self):
        pass



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Dict, Union, List, Type, Optional

from iconcommons.logger import Logger

from .base.exception import InternalServiceErrorException
from .icon_constant import (
    ConfigKey, TERM_PERIOD, IISS_DAY_BLOCK, PREP_MAIN_PREPS,
    PREP_MAIN_AND_SUB_PREPS, PENALTY_GRACE_PERIOD, LOW_PRODUCTIVITY_PENALTY_THRESHOLD,
    BLOCK_VALIDATION_PENALTY_THRESHOLD, BACKUP_FILES, BLOCK_INVOKE_TIMEOUT_S,
    IISS_INITIAL_IREP, PREP_REGISTRATION_FEE, UNSTAKE_SLOT_MAX)

_TAG = "CFG"
ConfigValue = Union[bool, dict, float, int, str]
ConfigKeyTypeDict = Dict[str, Union[dict, Type[Union[bool, float, int, str]]]]

default_icon_config = {
    ConfigKey.LOG: {
        ConfigKey.LOGGER: "iconservice",
        ConfigKey.LOG_FILE_PATH: "./",
        ConfigKey.LOG_LEVEL: "debug"
    },
    ConfigKey.SCORE_ROOT_PATH: ".score",
    ConfigKey.STATE_DB_ROOT_PATH: ".statedb",
    ConfigKey.CHANNEL: "loopchain_default",
    ConfigKey.AMQP_KEY: "7100",
    ConfigKey.AMQP_TARGET: "127.0.0.1",
    ConfigKey.BUILTIN_SCORE_OWNER: "hxebf3a409845cd09dcb5af31ed5be5e34e2af9433",
    ConfigKey.IPC_TIMEOUT: 10,
    ConfigKey.SERVICE: {
        ConfigKey.SERVICE_FEE: False,
        ConfigKey.SERVICE_AUDIT: False,
        ConfigKey.SERVICE_SCORE_PACKAGE_VALIDATOR: False
    },
    ConfigKey.IISS_META_DATA: {
        ConfigKey.REWARD_MIN: 200,
        ConfigKey.REWARD_MAX: 1200,
        ConfigKey.REWARD_POINT: 7000,
        ConfigKey.UN_STAKE_LOCK_MIN: IISS_DAY_BLOCK * 5,
        ConfigKey.UN_STAKE_LOCK_MAX: IISS_DAY_BLOCK * 20
    },
    # The reason why IISS_CALCULATE_PERIOD and TERM_PERIOD value is different
    # is to synchronize with the main net without revisioning.
    # In the main net, calculate period in prevote is 43_200,
    # and after decentralizing, is set to 43_120.
    # If you want to change as TERM_PERIOD, you also must change REVISION.
    # so we determined that only TERM_PERIOD changed without IISS_CALCULATE_PERIOD.
    ConfigKey.ICON_RC_DIR_PATH: "",
    ConfigKey.ICON_RC_MONITOR: True,
    ConfigKey.IISS_CALCULATE_PERIOD: IISS_DAY_BLOCK,
    ConfigKey.TERM_PERIOD: TERM_PERIOD,
    ConfigKey.INITIAL_IREP: IISS_INITIAL_IREP,
    ConfigKey.PREP_REGISTRATION_FEE: PREP_REGISTRATION_FEE,
    ConfigKey.PREP_MAIN_PREPS: PREP_MAIN_PREPS,
    ConfigKey.PREP_MAIN_AND_SUB_PREPS: PREP_MAIN_AND_SUB_PREPS,
    ConfigKey.DECENTRALIZE_TRIGGER: 0.002,
    ConfigKey.PENALTY_GRACE_PERIOD: PENALTY_GRACE_PERIOD,
    ConfigKey.LOW_PRODUCTIVITY_PENALTY_THRESHOLD: LOW_PRODUCTIVITY_PENALTY_THRESHOLD,
    ConfigKey.BLOCK_VALIDATION_PENALTY_THRESHOLD: BLOCK_VALIDATION_PENALTY_THRESHOLD,
    ConfigKey.STEP_TRACE_FLAG: False,
    ConfigKey.PRECOMMIT_DATA_LOG_FLAG: False,
    ConfigKey.BACKUP_FILES: BACKUP_FILES,
    ConfigKey.BLOCK_INVOKE_TIMEOUT: BLOCK_INVOKE_TIMEOUT_S,
    ConfigKey.TBEARS_MODE: False,
    ConfigKey.UNSTAKE_SLOT_MAX: UNSTAKE_SLOT_MAX,
}


def check_config(conf: dict, default_conf: dict) -> bool:
    checker = ConfigSanityChecker(default_conf)
    ret: bool = checker.run(conf)

    if len(checker.invalid_keys) > 0:
        Logger.info(tag=_TAG, msg="===== INVALID CONFIGURATION KEYS =====")
        for item in checker.invalid_keys:
            Logger.info(tag=_TAG, msg=f"{item}")

    if len(checker.invalid_values) > 0:
        Logger.error(tag=_TAG, msg="===== INVALID CONFIGURATION VALUE TYPES ===== ")
        for item in checker.invalid_values:
            Logger.error(tag=_TAG, msg=f"{item}")

    return ret


def args_to_dict(args) -> dict:
    """Convert args to dict
    :param args: command line arguments
    :return: conf in dict containing command line arguments
    """
    keys_to_remove = {"command", "foreground", "config"}
    args_dict = vars(args)
    ret = {}

    for key, value in args_dict.items():
        if key not in keys_to_remove and value is not None:
            ret[key] = value

    return ret


class ConfigSanityChecker(object):
    """Sanity check for iconservice configuration file
    """

    class Item(object):
        def __init__(self,
                     key: str,
                     value: ConfigValue,
                     value_type: Optional[Type[ConfigValue]]):
            self._key = key
            self._value = value
            self._value_type = value_type

        @property
        def key(self) -> str:
            return self._key

        @property
        def value(self) -> ConfigValue:
            return self._value

        @property
        def value_type(self) -> Optional[Type[ConfigValue]]:
            return self._value_type

        def __str__(self) -> str:
            return f"key={self._key} value={self._value} value_type={self._value_type}"

    def __init__(self, default_conf: Dict[str, ConfigValue]):
        """
        :param default_conf: dict containing default configuration values
        """

        self._invalid_keys = []
        self._invalid_values = []
        # dict containing config_key:value_type
        self._config_key_type_dict: ConfigKeyTypeDict = self._init_config_key_type_dict(default_conf)
        self._add_extra_value_types(self._config_key_type_dict)

    @classmethod
    def _init_config_key_type_dict(cls, default_conf: dict) -> ConfigKeyTypeDict:
        config_key_type_dict = copy.deepcopy(default_conf)

        def get_value_type(value):
            if isinstance(value, dict):
                _table = value
                for key in _table:
                    _table[key] = get_value_type(_table[key])
            elif isinstance(value, (bool, float, int, str)):
                value = type(value)
            else:
                raise InternalServiceErrorException(f"Invalid default config: {value}")

            return value

        return get_value_type(config_key_type_dict)

    @classmethod
    def _add_extra_value_types(cls, table: ConfigKeyTypeDict):
        # Add some value types which cannot be contained to default_conf
        table[ConfigKey.LOG] = {
            ConfigKey.LOGGER: str,
            ConfigKey.LOG_FILE_PATH: str,
            ConfigKey.LOG_LEVEL: str,
            ConfigKey.LOG_OUTPUT_TYPE: str,
            ConfigKey.LOG_ROTATE: {
                ConfigKey.LOG_ROTATE_TYPE: str,
                ConfigKey.LOG_ROTATE_PERIOD: str,
                ConfigKey.LOG_ROTATE_AT_TIME: int,
                ConfigKey.LOG_ROTATE_INTERVAL: int,
                ConfigKey.LOG_ROTATE_MAX_BYTES: int,
                ConfigKey.LOG_ROTATE_BACKUP_COUNT: int,
            },
        }

    @property
    def invalid_keys(self) -> List[Item]:
        return self._invalid_keys

    @property
    def invalid_values(self) -> List:
        return self._invalid_values

    def run(self, conf: Dict[str, ConfigValue]) -> bool:
        self._invalid_keys.clear()
        self._invalid_values.clear()

        self._check(
            conf,
            self._config_key_type_dict,
            self._invalid_keys,
            self._invalid_values
        )

        return len(self._invalid_values) == 0

    @classmethod
    def _check(cls,
               conf: Dict[str, ConfigValue],
               key_type_table: ConfigKeyTypeDict,
               invalid_keys: List[Item],
               invalid_values: List[Item]):
        for key, value in conf.items():
            try:
                # Check if key is valid
                if key not in key_type_table:
                    raise KeyError(key, value)

                # Check if value is valid
                value_type: Type = cls._get_value_type(key_type_table, key)
                if type(value) is not value_type:
                    raise ValueError(key, value, value_type)

                if isinstance(value, dict):
                    # If value is a dict type, call _check() recursively
                    cls._check(
                        conf=value,
                        key_type_table=key_type_table[key],
                        invalid_keys=invalid_keys,
                        invalid_values=invalid_values
                    )
            except KeyError as e:
                invalid_keys.append(cls.Item(key=e.args[0], value=e.args[1], value_type=None))
            except ValueError as e:
                invalid_values.append(cls.Item(key=e.args[0], value=e.args[1], value_type=e.args[2]))

    @classmethod
    def _get_value_type(cls, key_type_table: ConfigKeyTypeDict, key: str) -> Type[ConfigValue]:
        """Returns the proper value type for configuration key from key_type_table

        :param key_type_table: key:value_type table
        :param key: configuration key
        :return: value type of a given configuration key
        """
        value_type = key_type_table[key]
        if not isinstance(value_type, type):
            # If value_type is a dict object, set type(value_type) to value_type
            value_type = type(value_type)

        return value_type



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum, unique, Flag
from struct import Struct
from typing import TYPE_CHECKING, Union

from .base_part import BasePart
from ..base.address import AddressPrefix
from ..base.exception import InvalidParamsException, OutOfBalanceException
from ..icon_constant import DEFAULT_BYTE_SIZE, DATA_BYTE_ORDER, Revision
from ..utils import set_flag
from ..utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from iconservice.base.address import Address


@unique
class CoinPartVersion(IntEnum):
    STRUCT = 0
    MSG_PACK = 1


@unique
class CoinPartType(IntEnum):
    """CoinPartType Type
    """
    GENERAL = 0
    GENESIS = 1
    TREASURY = 2

    def __str__(self) -> str:
        return self.name


class CoinPartFlag(Flag):
    NONE = 0
    HAS_UNSTAKE = 1 # deprecated


class CoinPart(BasePart):
    """Account class
    Contains information of the account indicated by address.
    """

    # leveldb account value structure (bigendian, 36 bytes)
    # version(1) | type(1) | flags(1) | reserved(1) |
    # icx(DEFAULT_BYTE_SIZE)

    _VERSION = CoinPartVersion.MSG_PACK
    _STRUCT_PACKED_BYTES_SIZE = 36
    _STRUCT_FORMAT = Struct(f'>BBBx{DEFAULT_BYTE_SIZE}s')

    def __init__(self,
                 coin_part_type: 'CoinPartType' = CoinPartType.GENERAL,
                 flags: 'CoinPartFlag' = CoinPartFlag.NONE,
                 balance: int = 0,
                 is_first: bool = True):
        """Constructor
        """
        super().__init__()
        assert isinstance(coin_part_type, CoinPartType)
        # Encoded bytes format is different Between from Rev IISS to Rev FIX_COIN_PART_BYTES_ENCODING
        # First recorded account: type is encoded as ExtType
        # Previously recorded account: type is encoded as primitive type
        # To distinguish, use is_first flag. It is only used between from Rev IISS to Rev FIX_COIN_PART_BYTES_ENCODING
        self._is_first: bool = is_first

        self._type: 'CoinPartType' = coin_part_type
        self._flags: 'CoinPartFlag' = flags
        self._balance: int = balance

    def __str__(self):
        return f"type={self._type.name}, " \
               f"flag={self._flags.name}, " \
               f"balance={self._balance}"

    @staticmethod
    def make_key(address: 'Address') -> bytes:
        if address.prefix == AddressPrefix.EOA:
            return address.body
        else:
            return address.to_bytes_including_prefix()

    @property
    def type(self) -> 'CoinPartType':
        """CoinPartType getter

        :return: CoinPartType value
        """
        return self._type

    @type.setter
    def type(self, value: 'CoinPartType'):
        """CoinPartType setter

        :param value: (CoinPartType)
        """
        if not isinstance(value, CoinPartType):
            raise ValueError('Invalid CoinPartType')
        self._type = value

    @property
    def balance(self) -> int:
        """Returns the balance of the account in loop unit (1 icx == 1e18 loop)

        :return: balance in loop
        """
        return self._balance

    @property
    def flags(self) -> 'CoinPartFlag':
        """CoinPartFlag getter

        :return: CoinPartType value
        """
        return self._flags

    def toggle_has_unstake(self, on: bool, revision: int):
        # deprecated flag
        if revision >= Revision.FIX_BALANCE_BUG.value:
            return

        new_flags = set_flag(self._flags, CoinPartFlag.HAS_UNSTAKE, on)

        if self._flags != new_flags:
            self._flags = new_flags
            self.set_dirty(True)

    def deposit(self, value: int):
        """Deposit coin

        :param value: amount to deposit in loop (1 icx == 1e18 loop)

        """
        if not isinstance(value, int) or value < 0:
            raise InvalidParamsException(
                'Failed to deposit: value is not int type or value < 0')

        self._increase_balance(value)

    def withdraw(self, value: int):
        """Withdraw coin

        :param value: coin amount to withdraw
        """
        if not isinstance(value, int) or value < 0:
            raise InvalidParamsException(
                'Failed to withdraw: value is not int type or value < 0')

        if self._balance < value:
            raise OutOfBalanceException('Out of balance')

        self._increase_balance(-value)

    def _increase_balance(self, value: int):
        if value == 0:
            return

        self._balance += value
        self.set_dirty(True)

    def __eq__(self, other) -> bool:
        """operator == overriding

        :param other: (CoinPart)
        """
        return isinstance(other, CoinPart) \
               and self._balance == other._balance \
               and self._type == other._type \
               and self._flags == other._flags

    def __ne__(self, other) -> bool:
        """operator != overriding

        :param other: (CoinPart)
        """
        return not self.__eq__(other)

    @staticmethod
    def from_bytes(buf: bytes) -> 'CoinPart':
        """Create CoinPart object from bytes data

        :param buf: (bytes) bytes data including Account information
        :return: (CoinPart) account object
        """

        if len(buf) == CoinPart._STRUCT_PACKED_BYTES_SIZE and buf[0] == CoinPartVersion.STRUCT:
            return CoinPart._from_struct_packed_bytes(buf)
        else:
            return CoinPart._from_msg_packed_bytes(buf)

    @staticmethod
    def _from_struct_packed_bytes(buf: bytes) -> 'CoinPart':
        version, coin_type, flags, amount = CoinPart._STRUCT_FORMAT.unpack(buf)
        balance: int = int.from_bytes(amount, DATA_BYTE_ORDER)

        return CoinPart(coin_part_type=CoinPartType(coin_type),
                        flags=CoinPartFlag(flags),
                        balance=balance,
                        is_first=False)

    @staticmethod
    def _from_msg_packed_bytes(buf: bytes) -> 'CoinPart':
        data: list = MsgPackForDB.loads(buf)
        version: int = data[0]

        assert version <= CoinPart._VERSION

        if version != CoinPartVersion.MSG_PACK:
            raise InvalidParamsException(f"Invalid Account version: {version}")

        return CoinPart(coin_part_type=CoinPartType(data[1]),
                        flags=CoinPartFlag(data[2]),
                        balance=data[3],
                        is_first=False)

    def to_bytes(self, revision: int) -> bytes:
        """Convert CoinPart object to bytes

        :return: data including information of CoinPart object
        """
        if revision >= Revision.FIX_COIN_PART_BYTES_ENCODING.value:
            return self._to_fixed_msg_packed_bytes()
        elif revision >= Revision.IISS.value:
            return self._to_msg_packed_bytes()
        else:
            return self._to_struct_packed_bytes()

    def _to_fixed_msg_packed_bytes(self) -> bytes:
        data = [
            CoinPartVersion.MSG_PACK.value,
            self._type.value,
            self._flags.value,
            self.balance
        ]

        return MsgPackForDB.dumps(data)

    def _to_msg_packed_bytes(self) -> bytes:
        _type: Union[int, 'CoinPartType'] = self._type if self._is_first else self._type.value
        data = [
            CoinPartVersion.MSG_PACK,
            _type,
            self._flags.value,
            self.balance
        ]

        return MsgPackForDB.dumps(data)

    def _to_struct_packed_bytes(self) -> bytes:
        return CoinPart._STRUCT_FORMAT.pack(CoinPartVersion.STRUCT,
                                            self._type,
                                            self._flags.value,
                                            self._balance.to_bytes(DEFAULT_BYTE_SIZE, DATA_BYTE_ORDER))



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Optional, List, Dict, Tuple

from iconcommons.logger import Logger

from iconservice.iiss.listener import EngineListener as IISSEngineListener
from .reward_calc.data_creator import DataCreator as RewardCalcDataCreator
from .reward_calc.ipc.message import CalculateDoneNotification, ReadyNotification
from .reward_calc.ipc.reward_calc_proxy import RewardCalcProxy
from ..base.ComponentBase import EngineBase
from ..base.address import Address, SYSTEM_SCORE_ADDRESS
from ..base.block import Block
from ..base.exception import (
    InvalidParamsException, InvalidRequestException, OutOfBalanceException, FatalException,
    InternalServiceErrorException
)
from ..base.transaction import Transaction
from ..icon_constant import ISCORE_EXCHANGE_RATE, IISS_MAX_REWARD_RATE, \
    IconScoreContextType, IISS_LOG_TAG, ROLLBACK_LOG_TAG, RCCalculateResult, INVALID_CLAIM_TX, Revision, \
    RevisionChangedFlag
from ..iconscore.icon_score_context import IconScoreContext
from ..iconscore.icon_score_event_log import EventLogEmitter
from ..iconscore.icon_score_step import StepType
from ..iconscore.system_score import Delegation
from ..icx import Intent
from ..icx.icx_account import Account
from ..icx.issue.issue_formula import IssueFormula
from ..iiss.reward_calc.storage import get_rc_version
from ..utils import bytes_to_hex

if TYPE_CHECKING:
    from .reward_calc.msg_data import TxData, DelegationInfo, DelegationTx, Header, BlockProduceInfoData, PRepsData
    from .reward_calc.msg_data import GovernanceVariable
    from ..iiss.storage import RewardRate
    from ..icx import IcxStorage
    from ..prep.data import Term

_TAG = IISS_LOG_TAG

QUERY_CALCULATE_REPEAT_COUNT = 3


class Method:
    SET_STAKE = 'setStake'
    GET_STAKE = 'getStake'
    SET_DELEGATION = 'setDelegation'
    GET_DELEGATION = 'getDelegation'
    CLAIM_ISCORE = 'claimIScore'
    QUERY_ISCORE = 'queryIScore'
    ESTIMATE_UNLOCK_PERIOD = 'estimateUnstakeLockPeriod'


class Engine(EngineBase):
    """IISSEngine class

    """
    TAG = "IISS"

    INVOKE_METHOD_TABLE = [
        Method.SET_STAKE,
        Method.SET_DELEGATION,
        Method.CLAIM_ISCORE,
    ]
    QUERY_METHOD_TABLE = [
        Method.GET_STAKE,
        Method.GET_DELEGATION,
        Method.QUERY_ISCORE,
        Method.ESTIMATE_UNLOCK_PERIOD,
    ]
    METHOD_TABLE = INVOKE_METHOD_TABLE + QUERY_METHOD_TABLE

    def __init__(self):
        super().__init__()

        self._invoke_handlers: dict = {
            Method.SET_STAKE: self.handle_set_stake,
            Method.SET_DELEGATION: self.handle_set_delegation,
            Method.CLAIM_ISCORE: self.handle_claim_iscore,
            Method.GET_STAKE: self.handle_get_stake,
            Method.GET_DELEGATION: self.handle_get_delegation,
            Method.QUERY_ISCORE: self.handle_query_iscore,
            Method.ESTIMATE_UNLOCK_PERIOD: self.handle_estimate_unstake_lock_period,
        }

        self._query_handler: dict = {
            Method.GET_STAKE: self.handle_get_stake,
            Method.GET_DELEGATION: self.handle_get_delegation,
            Method.QUERY_ISCORE: self.handle_query_iscore,
            Method.ESTIMATE_UNLOCK_PERIOD: self.handle_estimate_unstake_lock_period,
        }

        self._reward_calc_proxy: Optional['RewardCalcProxy'] = None
        self._listeners: List['IISSEngineListener'] = []

    def open(self, context: 'IconScoreContext',
             log_dir: str, data_path: str, socket_path: str, ipc_timeout: int,
             icon_rc_path: str, icon_rc_monitor: bool):
        """
        :param context:
        :param log_dir:
        :param data_path:
        :param socket_path:
        :param ipc_timeout:
        :param icon_rc_path: ex) "/usr/local/bin"
        :param icon_rc_monitor: Boolean which determines Opening RC monitor channel
        :return:
        """
        self._init_reward_calc_proxy(log_dir, data_path, socket_path, ipc_timeout, icon_rc_path, icon_rc_monitor)

    def add_listener(self, listener: 'IISSEngineListener'):
        assert isinstance(listener, IISSEngineListener)
        self._listeners.append(listener)

    def remove_listener(self, listener: 'IISSEngineListener'):
        assert isinstance(listener, IISSEngineListener)
        self._listeners.remove(listener)

    @staticmethod
    def ready_callback(cb_data: 'ReadyNotification'):
        Logger.debug(tag=_TAG, msg=f"ready_callback() start: {cb_data}")
        Logger.debug(tag=_TAG, msg="ready_callback() end")

    def init_reward_calculator(self, block: 'Block'):
        """Send a INIT request to RC to synchronize the block state with RC

        """
        Logger.debug(tag=_TAG, msg=f"_init_reward_calculator() start: block={block}")

        success, block_height = self._reward_calc_proxy.init_reward_calculator(block.height)
        Logger.info(tag=_TAG, msg=f"success={success} block_height={block_height} last_block={block}")

        Logger.debug(tag=_TAG, msg=f"_init_reward_calculator() end")

    def get_ready_future(self):
        return self._reward_calc_proxy.get_ready_future()

    def is_reward_calculator_ready(self):
        return self._reward_calc_proxy.is_reward_calculator_ready()

    def query_calculate_result(self,
                               calc_bh: int,
                               repeat_cnt: int = QUERY_CALCULATE_REPEAT_COUNT) -> Tuple[int, int, bytes]:
        """Query the calculation result for the last term to reward calculator

        :param calc_bh:
        :param repeat_cnt: retry count
        :return:
        """
        Logger.debug(tag=_TAG, msg=f"_query_calculate_result() start")

        calc_result_status: int = -1
        calc_result_bh: int = -1
        state_hash: Optional[bytes] = None
        iscore: int = -1

        for i in range(repeat_cnt):
            calc_result_status, calc_result_bh, iscore, state_hash = \
                self._reward_calc_proxy.query_calculate_result(calc_bh)
            if calc_result_status == RCCalculateResult.SUCCESS:
                break
            elif calc_result_status == RCCalculateResult.IN_PROGRESS:
                time.sleep(1)
                Logger.debug(tag=_TAG, msg=f"Retry to query calculate result: {i + 1}/{repeat_cnt}")
                continue
            else:
                raise FatalException(f'RC has a problem about calculating: {calc_result_status}')

        if calc_result_status != RCCalculateResult.SUCCESS:
            raise FatalException(f'RC has a problem about calculating: {calc_result_status}')

        if calc_result_bh != calc_bh:
            raise FatalException(f'Unexpected calculate result response '
                                 f'(reward calc: {calc_result_bh} icon service: {calc_bh}')

        if iscore < 0:
            raise FatalException(f'Invalid I-SCORE value: {iscore}')

        Logger.debug(tag=_TAG, msg=f"query_calculate_result() end: "
                                   f"status={calc_result_status}, "
                                   f"calc_result_bh={calc_result_bh}, "
                                   f"iscore={iscore}, "
                                   f"state_hash={bytes_to_hex(state_hash)}")

        return iscore, calc_result_bh, state_hash

    @staticmethod
    def check_calculate_request_block_height(reward_calc_bh: int,
                                             icon_service_bh: int):

        if reward_calc_bh != icon_service_bh:
            raise FatalException(f"request block height is not matched: "
                                 f"response from RC:{reward_calc_bh} "
                                 f"request:{reward_calc_bh} ")

    def calculate_done_callback(self, cb_data: 'CalculateDoneNotification'):
        Logger.info(tag=_TAG, msg=f"calculate_done_callback start")
        # cb_data.success == False: RC has reset the state to before 'CALCULATE' request
        if not cb_data.success:
            raise FatalException(f"Reward calc has failed calculating about block height:{cb_data.block_height}")

        # context for searching db data
        context: 'IconScoreContext' = IconScoreContext(IconScoreContextType.QUERY)
        end_block_height_of_calc: int = context.storage.iiss.get_end_block_height_of_calc(context)
        calc_period: int = context.storage.iiss.get_calc_period(context)

        latest_calculate_bh: int = end_block_height_of_calc - calc_period
        self.check_calculate_request_block_height(cb_data.block_height, latest_calculate_bh)

        IconScoreContext.storage.rc.put_calc_response_from_rc(cb_data.iscore, cb_data.block_height, cb_data.state_hash)
        Logger.info(tag=_TAG, msg=f"calculate done callback called with {cb_data}")

    def _init_reward_calc_proxy(self, log_dir: str, data_path: str, socket_path: str, ipc_timeout: int,
                                icon_rc_path: str, icon_rc_monitor: bool):
        self._reward_calc_proxy = RewardCalcProxy(calc_done_callback=self.calculate_done_callback,
                                                  ready_callback=self.ready_callback,
                                                  ipc_timeout=ipc_timeout,
                                                  icon_rc_path=icon_rc_path)
        self._reward_calc_proxy.open(sock_path=socket_path)
        self._reward_calc_proxy.start(
            log_dir=log_dir,
            sock_path=socket_path,
            iiss_db_path=data_path,
            icon_rc_monitor=icon_rc_monitor
        )

    def _close_reward_calc_proxy(self):
        self._reward_calc_proxy.stop()
        self._reward_calc_proxy.close()

    def _init_handlers(self, handlers: list):
        for handler in handlers:
            handler.reward_calc_proxy = self._reward_calc_proxy

    def close(self):
        self._close_reward_calc_proxy()

    @classmethod
    def check_method(cls, method: str) -> bool:
        return method in cls.METHOD_TABLE

    @classmethod
    def check_invoke_method(cls, method: str) -> bool:
        return method in cls.INVOKE_METHOD_TABLE

    @classmethod
    def check_query_method(cls, method: str) -> bool:
        return method in cls.QUERY_METHOD_TABLE

    def invoke(self, context: 'IconScoreContext', method: str, params: dict):
        if context.revision < Revision.IISS.value:
            context.step_counter.apply_step(StepType.CONTRACT_CALL, 1)
            raise InvalidParamsException(f"Method Not Found: {method}")

        handler: callable = self._invoke_handlers[method]
        handler(context, **params)

    def query(self, context: 'IconScoreContext', method: str, params: dict) -> Any:
        if context.revision < Revision.SYSTEM_SCORE_ENABLED.value and \
                context.type == IconScoreContextType.INVOKE:
            raise InvalidRequestException(f"Do not call readonly method '{method}' with 'icx_sendTransaction'")
        handler: callable = self._query_handler[method]
        ret = handler(context, **params)
        return ret

    def handle_set_stake(self, context: 'IconScoreContext', value: int):
        if not isinstance(value, int) or value < 0:
            raise InvalidParamsException(
                "Failed to stake: value is not int type or value < 0"
            )

        address: 'Address' = context.msg.sender
        account: 'Account' = context.storage.icx.get_account(context, address, Intent.ALL)
        total_stake: int = context.storage.iiss.get_total_stake(context)

        if value < 0:
            raise InvalidParamsException('Failed to stake: value < 0')

        self._check_from_can_stake(context, value, account)

        unstake_lock_period: int = self._calculate_unstake_lock_period(context.storage.iiss.lock_min,
                                                                       context.storage.iiss.lock_max,
                                                                       context.storage.iiss.reward_point,
                                                                       total_stake,
                                                                       context.total_supply)
        # subtract account's staked amount from the total stake
        total_stake -= account.stake
        account.set_stake(context, value, unstake_lock_period)
        # add account's newly set staked amount from the total stake
        total_stake += account.stake
        context.storage.icx.put_account(context, account)
        context.storage.iiss.put_total_stake(context, total_stake)
        # TODO tx_result make if needs
        for listener in self._listeners:
            listener.on_set_stake(context, account)

    @classmethod
    def _check_from_can_stake(cls,
                              context: 'IconScoreContext',
                              stake: int,
                              account: 'Account'):
        if context.revision < Revision.SYSTEM_SCORE_ENABLED.value:
            fee = context.step_counter.step_price * context.step_counter.step_used
        else:
            # IconServiceEngine will check fee
            fee = 0

        if account.balance + account.total_stake < stake + fee:
            raise OutOfBalanceException(
                f'Out of balance({account.address}): balance({account.balance}) + total_stake({account.total_stake})'
                f' < stake({stake}) + fee({fee})')

        if stake < account.delegations_amount:
            raise InvalidParamsException(f"Failed to stake: stake({stake})"
                                         f" < delegations_amount({account.delegations_amount})")

    @classmethod
    def _calculate_unstake_lock_period(cls,
                                       lmin: int,
                                       lmax: int,
                                       rpoint: int,
                                       total_stake: int,
                                       total_supply: int):
        stake_percentage: float = total_stake / total_supply
        if stake_percentage >= rpoint / IISS_MAX_REWARD_RATE:
            return lmin

        first_operand: float = (lmax - lmin) / (rpoint / IISS_MAX_REWARD_RATE) ** 2
        second_operand: float = (stake_percentage - (rpoint / IISS_MAX_REWARD_RATE)) ** 2
        return int(first_operand * second_operand) + lmin

    @staticmethod
    def handle_get_stake(context: 'IconScoreContext', address: "Address") -> dict:
        account: "Account" = context.storage.icx.get_account(
            context, address, Intent.STAKE
        )

        stake: int = account.stake
        unstake: int = account.unstake
        unstake_block_height: int = account.unstake_block_height
        unstakes_info: Optional[List[Tuple[int, int]]] = account.unstakes_info

        data = {
            "stake": stake
        }

        if context.revision < Revision.MULTIPLE_UNSTAKE.value:
            if unstake_block_height:
                data["unstake"] = unstake
                data["unstakeBlockHeight"] = unstake_block_height
                data["remainingBlocks"] = unstake_block_height - context.block.height

        elif context.revision >= Revision.MULTIPLE_UNSTAKE.value:
            if unstakes_info:
                data["unstakes"] = [
                    {
                        "unstake": unstakes_data[0],
                        "unstakeBlockHeight": unstakes_data[1],
                        "remainingBlocks": unstakes_data[1] - context.block.height
                     }
                    for unstakes_data in unstakes_info
                ]
        return data

    def handle_estimate_unstake_lock_period(self, context: 'IconScoreContext'):
        total_stake: int = context.storage.iiss.get_total_stake(context)
        unstake_lock_period: int = self._calculate_unstake_lock_period(context.storage.iiss.lock_min,
                                                                       context.storage.iiss.lock_max,
                                                                       context.storage.iiss.reward_point,
                                                                       total_stake,
                                                                       context.total_supply)
        return {
            "unstakeLockPeriod": unstake_lock_period
        }

    def handle_set_delegation(self, context: 'IconScoreContext', delegations: List[Delegation]):
        """Handles setDelegation JSON-RPC API request
        """
        # SCORE can stake via SCORE inter-call
        sender: 'Address' = context.msg.sender
        cached_accounts: Dict['Address', Tuple['Account', int]] = OrderedDict()

        # Convert setDelegation params
        total_delegating, new_delegations = \
            self._convert_params_of_set_delegation(context, delegations)

        # Check whether voting power is enough to delegate
        self._check_voting_power_is_enough(context, sender, total_delegating, cached_accounts)

        # Get old delegations from delegating accounts
        self._get_old_delegations_from_sender_account(context, sender, cached_accounts)

        # Calculate delegations with old and new delegations
        self._calc_delegations(context, new_delegations, cached_accounts)

        # Put updated delegation data to stateDB
        updated_accounts: List['Account'] = \
            self._put_delegation_to_state_db(context, sender, new_delegations, cached_accounts)

        # Put updated delegation data to rcDB
        self._put_delegation_to_rc_db(context, sender, new_delegations)

        for listener in self._listeners:
            listener.on_set_delegation(context, updated_accounts)

    @classmethod
    def get_max_delegations_by_revision(cls, context: 'IconScoreContext') -> int:
        if context.revision >= Revision.CHANGE_MAX_DELEGATIONS_TO_100.value:
            max_delegations: int = 100
        else:
            # Initial max delegations
            max_delegations: int = 10
        return max_delegations

    @classmethod
    def _check_delegation_count(cls,
                                context: 'IconScoreContext',
                                delegations: List):
        assert isinstance(delegations, list)

        if len(delegations) > cls.get_max_delegations_by_revision(context):
            raise InvalidParamsException(f"Delegations out of range: {len(delegations)}")

    @classmethod
    def _convert_params_of_set_delegation(cls,
                                          context: 'IconScoreContext',
                                          delegations: Optional[List[Delegation]]
                                          ) -> Tuple[int, List[Tuple['Address', int]]]:
        """Convert delegations format

        [{"address": Address(hxe7af5fcfd8dfc67530a01a0e403882687528dfcb), "value", 1234}, ...] ->
        [(Address(hxe7af5fcfd8dfc67530a01a0e403882687528dfcb), 1234), ...]

        :param delegations: delegations of setDelegation JSON-RPC API request
        :return: total_delegating, (address, delegated)
        """
        assert delegations is None or isinstance(delegations, list)

        if delegations is None or len(delegations) == 0:
            return 0, []

        cls._check_delegation_count(context, delegations)

        total_delegating: int = 0
        converted_delegations: List[Tuple['Address', int]] = []
        delegated_addresses = set()

        for delegation in delegations:
            address: 'Address' = delegation["address"]
            value: int = delegation["value"]
            assert isinstance(address, Address)
            assert isinstance(value, int)

            if value < 0:
                raise InvalidParamsException(f"Invalid delegating amount: {value}")

            if address in delegated_addresses:
                raise InvalidParamsException(f"Duplicated address: {address}")

            delegated_addresses.add(address)

            if value > 0:
                total_delegating += value
                converted_delegations.append((address, value))

        return total_delegating, converted_delegations

    @classmethod
    def _check_voting_power_is_enough(cls,
                                      context: 'IconScoreContext',
                                      sender: 'Address', delegating: int,
                                      cached_accounts: Dict['Address', Tuple['Account', int]]):
        """

        :param context:
        :param sender:
        :param cached_accounts:
        :param delegating:
        :return:
        """
        account: 'Account' = context.storage.icx.get_account(context, sender, Intent.ALL)
        assert isinstance(account, Account)

        if account.voting_weight < delegating:
            raise InvalidRequestException("Not enough voting power")

        cached_accounts[sender] = account, 0

    @classmethod
    def _get_old_delegations_from_sender_account(cls,
                                                 context: 'IconScoreContext',
                                                 sender: 'Address',
                                                 cached_accounts: Dict['Address', Tuple['Account', int]]):
        """Get old delegations from sender account

        :param context:
        :param sender:
        :param cached_accounts:
        :return:
        """
        icx_storage = context.storage.icx
        sender_account: 'Account' = cached_accounts[sender][0]

        old_delegations: List[Tuple[Address, int]] = sender_account.delegations
        if not old_delegations:
            return

        for address, old_delegated in old_delegations:
            assert old_delegated > 0

            cached: Tuple['Account', int] = cached_accounts.get(address)
            if cached is None:
                account: 'Account' = icx_storage.get_account(context, address, Intent.DELEGATED)
            else:
                account: 'Account' = cached[0]

            assert account.delegated_amount >= old_delegated
            cached_accounts[address] = account, -old_delegated

    @classmethod
    def _calc_delegations(cls,
                          context: 'IconScoreContext',
                          new_delegations: List[Tuple['Address', int]],
                          cached_accounts: Dict['Address', Tuple['Account', int]]):
        """Calculate new delegated amounts for each address with old and new delegations

        :param new_delegations:
        :param cached_accounts:
        :return:
        """
        icx_storage = context.storage.icx

        for address, new_delegated in new_delegations:
            assert new_delegated > 0

            if address in cached_accounts:
                account, old_delegated = cached_accounts[address]
            else:
                account: 'Account' = icx_storage.get_account(context, address, Intent.DELEGATED)
                old_delegated = 0

            cached_accounts[address] = account, new_delegated + old_delegated

    @classmethod
    def _put_delegation_to_state_db(cls,
                                    context: 'IconScoreContext',
                                    sender: 'Address',
                                    delegations: List[Tuple['Address', int]],
                                    cached_accounts: Dict['Address', Tuple['Account', int]]) -> List['Account']:
        """Put updated delegations to stateDB

        :param context:
        :param sender:
        :param delegations:
        :param cached_accounts:
        :return: updated account list
        """
        cls._check_delegation_count(context, delegations)

        icx_storage: 'IcxStorage' = context.storage.icx

        sender_account: 'Account' = cached_accounts[sender][0]
        sender_account.set_delegations(delegations)

        updated_accounts: List['Account'] = []

        # Save changed accounts to stateDB
        for account, delegated_offset in cached_accounts.values():
            if delegated_offset != 0:
                account.delegation_part.delegated_amount += delegated_offset

            icx_storage.put_account(context, account)
            updated_accounts.append(account)

        return updated_accounts

    @classmethod
    def _put_delegation_to_rc_db(cls,
                                 context: 'IconScoreContext',
                                 address: 'Address',
                                 delegations: List[Tuple['Address', int]]):
        """Put new delegations from setDelegation JSON-RPC API request to RewardCalcDB

        :param context:
        :param address: The address of delegating account
        :param delegations: The list of delegations that the given address did
        :return:
        """
        delegation_list: list = []

        for delegation in delegations:
            info: 'DelegationInfo' = RewardCalcDataCreator.create_delegation_info(*delegation)
            delegation_list.append(info)

        delegation_tx: 'DelegationTx' = RewardCalcDataCreator.create_tx_delegation(delegation_list)
        iiss_tx_data: 'TxData' = RewardCalcDataCreator.create_tx(address, context.block.height, delegation_tx)
        context.storage.rc.put(context.rc_block_batch, iiss_tx_data)

    def handle_get_delegation(self, context: 'IconScoreContext', address: 'Address') -> dict:
        """Handles getDelegation JSON-RPC API request
        """
        account: 'Account' = context.storage.icx.get_account(context, address, Intent.ALL)
        delegation_list: list = []
        for address, value in account.delegations:
            delegation_list.append({"address": address, "value": value})

        data = {
            "delegations": delegation_list,
            "totalDelegated": account.delegations_amount,
            "votingPower": account.voting_power
        }

        return data

    @classmethod
    def _iscore_to_icx(cls,
                       iscore: int) -> int:
        """Exchange iscore to icx

        10 ** -18 icx == 1 loop == 1000 iscore

        :param iscore:
        :return: icx
        """
        return iscore // ISCORE_EXCHANGE_RATE

    def handle_claim_iscore(self, context: 'IconScoreContext'):
        """Handles claimIScore JSON-RPC request
        """
        address: Address = context.msg.sender
        iscore, block_height = self._claim_iscore(context, address)

        if iscore > 0:
            self._commit_claim(context, iscore, address)

        if context.revision < Revision.SYSTEM_SCORE_ENABLED.value:
            EventLogEmitter.emit_event_log(
                context,
                score_address=SYSTEM_SCORE_ADDRESS,
                event_signature="IScoreClaimed(int,int)",
                arguments=[iscore, self._iscore_to_icx(iscore)],
                indexed_args_count=0
            )
        else:
            EventLogEmitter.emit_event_log(
                context,
                score_address=SYSTEM_SCORE_ADDRESS,
                event_signature="IScoreClaimedV2(Address,int,int)",
                arguments=[address, iscore, self._iscore_to_icx(iscore)],
                indexed_args_count=1
            )

    @staticmethod
    def _check_claim_tx(context: 'IconScoreContext') -> bool:
        if context.tx.hash in INVALID_CLAIM_TX:
            Logger.error(tag=_TAG, msg=f"skip claim tx: {context.tx.hash.hex()}")
            return False
        else:
            return True

    def _claim_iscore(self, context: 'IconScoreContext', address: Address) -> (int, int):
        block: 'Block' = context.block
        tx: 'Transaction' = context.tx

        if context.type == IconScoreContextType.INVOKE and self._check_claim_tx(context):
            iscore, block_height = self._reward_calc_proxy.claim_iscore(
                address, block.height, block.hash, tx.index, tx.hash)
        else:
            # For debug_estimateStep request
            iscore, block_height = 0, 0

        return iscore, block_height

    def _commit_claim(self, context: 'IconScoreContext', iscore: int, address: Address):
        block: 'Block' = context.block
        tx: 'Transaction' = context.tx
        success = True

        try:
            icx: int = self._iscore_to_icx(iscore)

            from_account: 'Account' = context.storage.icx.get_account(context, address)
            treasury_address: 'Address' = context.storage.icx.fee_treasury
            treasury_account: 'Account' = context.storage.icx.get_account(context, treasury_address)

            treasury_account.withdraw(icx)
            from_account.deposit(icx)
            context.storage.icx.put_account(context, treasury_account)
            context.storage.icx.put_account(context, from_account)

        except BaseException as e:
            Logger.exception(tag=_TAG, msg=str(e))
            success = False
            raise e
        finally:
            self._reward_calc_proxy.commit_claim(success, address, block.height, block.hash, tx.index, tx.hash)

    def handle_query_iscore(self, context: 'IconScoreContext', address: 'Address') -> dict:
        if not isinstance(address, Address):
            raise InvalidParamsException(f"Invalid address: {address}")

        tx_hash = context.tx.hash if isinstance(context.tx, Transaction) else None
        block = context.block if isinstance(context.block, Block) else None
        iscore, block_height = self._reward_calc_proxy.query_iscore(address, block, tx_hash)

        data = {
            "iscore": iscore,
            "estimatedICX": self._iscore_to_icx(iscore),
            "blockHeight": block_height
        }

        return data

    def update_db(self,
                  context: 'IconScoreContext',
                  term: Optional['Term'],
                  prev_block_generator: Optional['Address'],
                  prev_block_votes: Optional[List[Tuple['Address', int]]],
                  rc_db_revision: int) -> Optional[bytes]:
        """Called on IconServiceEngine._after_transaction_process()

        :param context:
        :param term:
        :param prev_block_generator:
        :param prev_block_votes:
        :param rc_db_revision:
        :return: rc_state_hash
        """
        version: int = get_rc_version(rc_db_revision)

        rc_state_hash: Optional[bytes] = None
        if self._is_iiss_calc(context.revision_changed_flag):
            self._update_state_db_on_end_calc(context)
            if bool(context.revision_changed_flag & RevisionChangedFlag.GENESIS_IISS_CALC):
                self._put_header_to_rc_db(context, context.revision, 0, is_genesis_iiss=True)

            # get rc_state_hash in calc done response.
            _, _, rc_state_hash = context.storage.rc.get_calc_response_from_rc()

        start: int = self.get_start_block_of_calc(context)
        # New calculation period is started
        if start == context.block.height:
            self._put_header_to_rc_db(context, rc_db_revision, version)
            self._put_gv_to_rc_db(context, version)

        if not context.is_decentralized():
            return rc_state_hash

        if not context.is_the_first_block_on_decentralization():
            self.put_block_produce_info_to_rc_db(context,
                                                 context.rc_block_batch,
                                                 prev_block_generator,
                                                 prev_block_votes)

        start_term_block: int = context.engine.prep.term.start_block_height
        # New P-Rep Term is started
        if start_term_block == context.block.height:
            self._put_preps_to_rc_db(context, context.revision)
            self._put_gv_to_rc_db(context, version)

        if term is not None and term.is_in_term(context.block.height):
            self._put_preps_to_rc_db(context, context.revision, term)

        return rc_state_hash

    def _update_state_db_on_end_calc(self, context: 'IconScoreContext'):
        # Warning: do not change the order of putting data (for state sync)
        self._put_last_calc_info(context)
        self._put_end_calc_block_height(context)
        self._put_rrep(context)

    def send_commit(self, block_height: int, block_hash: bytes):
        self._reward_calc_proxy.commit_block(True, block_height, block_hash)

    def send_calculate(self, iiss_db_path: str, block_height: int):
        self._reward_calc_proxy.calculate(iiss_db_path, block_height)

    @classmethod
    def _is_iiss_calc(cls,
                      flag: 'RevisionChangedFlag') -> bool:
        return bool(flag & (RevisionChangedFlag.GENESIS_IISS_CALC | RevisionChangedFlag.IISS_CALC))

    @classmethod
    def _check_update_calc_period(cls,
                                  context: 'IconScoreContext') -> bool:
        block_height: int = context.block.height
        check_end_block_height: Optional[int] = context.storage.iiss.get_end_block_height_of_calc(context)
        if check_end_block_height is None:
            return False

        return block_height == check_end_block_height

    @classmethod
    def _put_last_calc_info(cls,
                            context: 'IconScoreContext'):

        _, last_calc_end = context.storage.meta.get_last_calc_info(context)
        if last_calc_end > 0:
            start: int = last_calc_end + 1
            end: int = context.block.height
        else:
            # first
            start: int = -1
            end: int = context.block.height
        context.storage.meta.put_last_calc_info(context,
                                                start,
                                                end)

    @classmethod
    def _put_end_calc_block_height(cls,
                                   context: 'IconScoreContext'):
        calc_period: int = context.storage.iiss.get_calc_period(context)
        if calc_period is None:
            raise InvalidParamsException("Fail put next calc block height: didn't init yet")
        context.storage.iiss.put_end_block_height_of_calc(context, context.block.height + calc_period)

    @classmethod
    def _put_header_to_rc_db(cls,
                             context: 'IconScoreContext',
                             rc_db_revision: int,
                             version: int,
                             is_genesis_iiss: bool = False):

        if is_genesis_iiss:
            block_height: int = context.block.height
        else:
            block_height: int = context.storage.iiss.get_end_block_height_of_calc(context)
        data: 'Header' = RewardCalcDataCreator.create_header(version,
                                                             block_height,
                                                             rc_db_revision)
        context.storage.rc.put(context.rc_block_batch, data)

    @staticmethod
    def _put_rrep(context: 'IconScoreContext'):
        reward_rate: 'RewardRate' = context.storage.iiss.get_reward_rate(context)
        reward_rate.reward_prep = IssueFormula.calculate_rrep(context.storage.iiss.reward_min,
                                                              context.storage.iiss.reward_max,
                                                              context.storage.iiss.reward_point,
                                                              context.storage.icx.get_total_supply(context),
                                                              context.preps.total_delegated)

        context.storage.iiss.put_reward_rate(context, reward_rate)

    @classmethod
    def _put_gv_to_rc_db(cls,
                         context: 'IconScoreContext',
                         version: int):

        calculated_irep: int = 0
        if context.is_decentralized():
            irep: int = context.engine.prep.term.irep
            calculated_irep: int = IssueFormula.calculate_irep_per_block_contributor(irep)
        reward_rate: 'RewardRate' = context.storage.iiss.get_reward_rate(context)
        reward_prep_for_rc = IssueFormula.calculate_temporary_reward_prep(reward_rate.reward_prep)

        # block height which GV variable has been calculated
        block_height: int = context.block.height - 1
        data: 'GovernanceVariable' = RewardCalcDataCreator.create_gv_variable(version,
                                                                              block_height,
                                                                              calculated_irep,
                                                                              reward_prep_for_rc,
                                                                              context.main_prep_count,
                                                                              context.main_and_sub_prep_count)
        context.storage.rc.put(context.rc_block_batch, data)

    @classmethod
    def put_block_produce_info_to_rc_db(cls,
                                        context: 'IconScoreContext',
                                        rc_block_batch: list,
                                        prev_block_generator: Optional['Address'] = None,
                                        prev_block_votes: Optional[List[Tuple['Address', int]]] = None):
        """Called on every block

        :param context:
        :param rc_block_batch:
        :param prev_block_generator:
        :param prev_block_votes:
        :return:
        """
        assert context.is_decentralized() and not context.is_the_first_block_on_decentralization()
        if prev_block_generator is None or prev_block_votes is None:
            return

        # Logger.debug(f"put_block_produce_info_for_rc", "IISS")
        prev_block_height: int = context.block.height - 1
        data: 'BlockProduceInfoData' = RewardCalcDataCreator.create_block_produce_info_data(prev_block_height,
                                                                                            prev_block_generator,
                                                                                            prev_block_votes)
        context.storage.rc.put(rc_block_batch, data)

    @classmethod
    def _put_preps_to_rc_db(cls, context: 'IconScoreContext', revision: int, term: Optional['Term'] = None):
        # If term is not None, it is the term which has been changed in term
        assert context.is_decentralized()

        if term is None:
            block_height: int = context.block.height - 1
            term: 'Term' = context.engine.prep.term
        else:
            block_height: int = context.block.height

        if revision < Revision.FIX_TOTAL_ELECTED_PREP_DELEGATED.value:
            total_elected_prep_delegated: int = term.total_elected_prep_delegated_snapshot
        else:
            total_elected_prep_delegated: int = term.total_elected_prep_delegated

        Logger.info(
            tag=cls.TAG,
            msg=f"_put_preps_for_rc_db() "
                f"block_height={block_height} "
                f"total_elected_prep_delegated={term.total_elected_prep_delegated} "
                f"total_elected_prep_delegated_snapshot={term.total_elected_prep_delegated_snapshot}")

        data: 'PRepsData' = RewardCalcDataCreator.create_prep_data(block_height,
                                                                   total_elected_prep_delegated,
                                                                   term.preps)
        context.storage.rc.put(context.rc_block_batch, data)

    @classmethod
    def get_start_block_of_calc(cls, context: 'IconScoreContext') -> int:
        start_calc_block: int = -1
        end_block_height: Optional[int] = context.storage.iiss.get_end_block_height_of_calc(context)
        period: Optional[int] = context.storage.iiss.get_calc_period(context)
        if end_block_height is not None and period is not None:
            start_calc_block: int = end_block_height - period + 1
        return start_calc_block

    def rollback_reward_calculator(self, block_height: int, block_hash: bytes):
        Logger.info(tag=ROLLBACK_LOG_TAG,
                    msg=f"rollback_reward_calculator() start: "
                        f"height={block_height} hash={bytes_to_hex(block_hash)}")

        _success, _height, _hash = self._reward_calc_proxy.rollback(block_height, block_hash)
        Logger.info(tag=ROLLBACK_LOG_TAG,
                    msg=f"RewardCalculator response: "
                        f"success={_success} height={_height} hash={bytes_to_hex(_hash)}")

        # Reward calculator rollback succeeded
        if _success and _height == block_height and _hash == block_hash:
            Logger.info(tag=ROLLBACK_LOG_TAG, msg=f"rollback_reward_calculator() end")
            return

        raise InternalServiceErrorException("Failed to rollback RewardCalculator")

    def get_reward_calculator_commit_block(self) -> Optional[Tuple[int, bytes]]:
        return self._reward_calc_proxy.get_commit_block()

    def send_start_block(self, block_height: int, block_hash: bytes):
        ret_block_height, ret_block_hash = \
            self._reward_calc_proxy.start_block(block_height, block_hash)
        if block_height == ret_block_height and block_hash == ret_block_hash:
            return

        raise InternalServiceErrorException("Failed to send start_block to reward_calculator")



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import TYPE_CHECKING, Optional, Tuple, Any, Dict

from .icon_score_context_util import IconScoreContextUtil
from ..base.exception import AccessDeniedException, IconServiceBaseException,\
    InvalidParamsException, InvalidRequestException
from ..icon_constant import IconNetworkValueType, Revision
from ..inv.container import ValueConverter as INVConverter
from ..inv.data.value import Value
from ..iconscore.icon_score_base import IconScoreBase
from ..prep.validator import validate_np_irep
from ..utils import is_builtin_score as util_is_builtin_score

if TYPE_CHECKING:
    from ..base.address import Address
    from ..database.db import IconScoreDatabase
    from ..deploy.storage import IconScoreDeployInfo, IconScoreDeployTXParams


class IconSystemScoreBase(IconScoreBase):

    @abstractmethod
    def on_install(self, **kwargs) -> None:
        """DB initialization on score install
        """
        super().on_install(**kwargs)

    @abstractmethod
    def on_update(self, **kwargs) -> None:
        """DB initialization on score update
        """
        super().on_update(**kwargs)

    @abstractmethod
    def __init__(self, db: 'IconScoreDatabase') -> None:
        super().__init__(db)
        if not util_is_builtin_score(str(self.address)):
            raise AccessDeniedException(f"Not a system SCORE ({self.address})")

    def is_builtin_score(self, score_address: 'Address') -> bool:
        return util_is_builtin_score(str(score_address))

    # TODO remove after Update 0.0.6
    def get_icon_service_flag(self) -> int:
        return self._context.icon_service_flag

    # TODO remove after Update 0.0.6
    def deploy(self, tx_hash: bytes) -> None:
        return IconScoreContextUtil.deploy(self._context, tx_hash)

    def get_deploy_tx_params(self, tx_hash: bytes) -> Optional['IconScoreDeployTXParams']:
        return IconScoreContextUtil.get_deploy_tx_params(self._context, tx_hash)

    def get_deploy_info(self, address: 'Address') -> Optional['IconScoreDeployInfo']:
        return IconScoreContextUtil.get_deploy_info(self._context, address)

    def is_score_active(self, score_address: 'Address') -> bool:
        return IconScoreContextUtil.is_score_active(self._context, score_address)

    def get_owner(self, score_address: Optional['Address']) -> Optional['Address']:
        return IconScoreContextUtil.get_owner(self._context, score_address)

    def migrate_icon_network_value(self, data: Dict['IconNetworkValueType', Any]):
        converted_data: list = []
        for type_, value in data.items():
            converted_data.append(INVConverter.convert_for_icon_service(type_, value))
        self._context.inv_container.migrate(converted_data)
        self._context.storage.inv.migrate(self._context, converted_data)

    def validate_irep(self, irep: int):
        if self._context.revision < Revision.SET_IREP_VIA_NETWORK_PROPOSAL.value:
            raise InvalidRequestException(f"Can't register I-Rep proposal. Revision must be larger than "
                                          f"{Revision.SET_IREP_VIA_NETWORK_PROPOSAL.value - 1}")

        validate_np_irep(self._context, irep)

    @classmethod
    def _check_inv_type(cls, type_: 'IconNetworkValueType'):
        if type_ not in IconNetworkValueType:
            raise InvalidParamsException(f"Invalid INV type: {type_}")

    def get_icon_network_value(self, type_: 'IconNetworkValueType') -> Any:
        self._check_inv_type(type_)

        value: Any = self._context.inv_container.get_by_type(type_)
        converted_value: Any = INVConverter.convert_for_governance(type_, value)
        return converted_value

    def set_icon_network_value(self, type_: 'IconNetworkValueType', value: Any):
        self._check_inv_type(type_)

        converted_value: 'Value' = INVConverter.convert_for_icon_service(type_, value)
        self._context.inv_container.set_inv_to_tx_batch(converted_value)
        self._context.storage.inv.put_value(self._context, converted_value)

    def disqualify_prep(self, address: 'Address') -> Tuple[bool, str]:
        success: bool = True
        reason: str = ""
        try:
            self._context.engine.prep.impose_prep_disqualified_penalty(self._context, address)
        except IconServiceBaseException as e:
            success = False
            reason = str(e)
        finally:
            return success, reason

    def apply_revision_change(self, revision):
        # CAUTION
        # Please follow the instructions below to apply the change. revision may not be set sequentially
        #   1. add an 'if' statement for every revision and compare revision with '>=' operator
        #   2. Check if the change has already been applied and apply the change

        if revision >= Revision.SET_IREP_VIA_NETWORK_PROPOSAL.value:
            # I-Rep is managed by INV. copy Term.irep to INV.irep
            if self.get_icon_network_value(IconNetworkValueType.IREP) is None:
                self.set_icon_network_value(IconNetworkValueType.IREP, self._context.term.irep)



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import TYPE_CHECKING

import iso3166

from ..base.exception import InvalidParamsException, InvalidRequestException
from ..base.type_converter_templates import ConstantKeys
from ..icon_constant import IISS_MIN_IREP, IISS_MAX_IREP_PERCENTAGE, IISS_MONTH, \
    PERCENTAGE_FOR_BETA_2, Revision

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext
    from .data import PRep, Term
    from ..base.address import Address

scheme_pattern = r'^(http:\/\/|https:\/\/)'
path_pattern = r'(\/\S*)?$'
port_regex = r'(:[0-9]{1,5})?'
ip_regex = r'(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])'
host_name_regex = r'(localhost|(?:[\w\d](?:[\w\d-]{0,61}[\w\d])\.)+[\w\d][\w\d-]{0,61}[\w\d])'
email_regex = r'^[a-zA-Z0-9]+(\.[a-zA-Z0-9]+)*@' + host_name_regex + '$'
ENDPOINT_DOMAIN_NAME_PATTERN = re.compile(f'^{host_name_regex}{port_regex}$')
ENDPOINT_IP_PATTERN = re.compile(f'^{ip_regex}{port_regex}$')
WEBSITE_DOMAIN_NAME_PATTERN = re.compile(f'{scheme_pattern}{host_name_regex}{port_regex}{path_pattern}$')
WEBSITE_IP_PATTERN = re.compile(f'{scheme_pattern}{ip_regex}{port_regex}{path_pattern}$')
EMAIL_PATTERN = re.compile(email_regex)
EMAIL_LOCAL_PART_MAX = 64
EMAIL_MAX = 254


def validate_prep_data(context: 'IconScoreContext',
                       prep_address: 'Address',
                       tx_data: dict,
                       set_prep: bool = False):
    if not set_prep:
        fields_to_validate = (
            ConstantKeys.NAME,
            ConstantKeys.COUNTRY,
            ConstantKeys.CITY,
            ConstantKeys.EMAIL,
            ConstantKeys.WEBSITE,
            ConstantKeys.DETAILS,
            ConstantKeys.P2P_ENDPOINT,
        )

        for key in fields_to_validate:
            if key not in tx_data:
                raise InvalidParamsException(f'"{key}" not found')
            elif isinstance(tx_data[key], str) and len(tx_data[key].strip()) < 1:
                raise InvalidParamsException("Empty data not allowed")

    for key, value in tx_data.items():
        if value is None:
            continue
        if isinstance(value, str) and len(value.strip()) < 1:
            raise InvalidParamsException("Empty data not allowed")
        if key == ConstantKeys.P2P_ENDPOINT:
            _validate_p2p_endpoint(context, value)
        elif key in (ConstantKeys.WEBSITE, ConstantKeys.DETAILS):
            _validate_uri(value)
        elif key == ConstantKeys.EMAIL:
            _validate_email(context.revision, value)
        elif key == ConstantKeys.COUNTRY:
            _validate_country(value)

    # node address validate
    is_update_node_address: bool = \
        ConstantKeys.NODE_ADDRESS in tx_data and tx_data[ConstantKeys.NODE_ADDRESS] is not None
    if set_prep and not is_update_node_address:
        # non changed skip
        return

    # register_prep or is_update_node_address is True
    node_address: 'Address' = tx_data[ConstantKeys.NODE_ADDRESS] if is_update_node_address else prep_address
    context.prep_address_converter.validate_node_address(node_address)


def _validate_p2p_endpoint(context: "IconScoreContext", p2p_endpoint: str):
    network_locate_info = p2p_endpoint.split(":")

    if len(network_locate_info) != 2:
        raise InvalidParamsException("Invalid endpoint format. endpoint must have port info")

    _validate_port(network_locate_info[1], ConstantKeys.P2P_ENDPOINT)

    if not (ENDPOINT_IP_PATTERN.match(p2p_endpoint) or ENDPOINT_DOMAIN_NAME_PATTERN.match(p2p_endpoint.lower())):
        raise InvalidParamsException("Invalid endpoint format")

    if context.revision >= Revision.PREVENT_DUPLICATED_ENDPOINT.value:
        for active_prep in context.preps:
            if active_prep.p2p_endpoint == p2p_endpoint and context.tx.origin != active_prep.address:
                raise InvalidParamsException("Duplicated endpoint")


def _validate_uri(uri: str):
    uri = uri.lower()
    if WEBSITE_DOMAIN_NAME_PATTERN.match(uri):
        return
    if WEBSITE_IP_PATTERN.match(uri):
        return

    raise InvalidParamsException("Invalid uri format")


def _validate_port(port: str, validating_field: str):
    try:
        port = int(port, 10)
    except ValueError:
        raise InvalidParamsException(f'Invalid {validating_field} format. port: "{port}"')

    if not 0 < port < 65536:
        raise InvalidParamsException(f"Invalid {validating_field} format. Port out of range: {port}")


def _validate_email(revision: int, email: str):
    error_msg = "Invalid email format"
    if revision < Revision.FIX_EMAIL_VALIDATION.value:
        if not EMAIL_PATTERN.match(email):
            raise InvalidParamsException(error_msg)
    else:
        encoded_email = email.encode('utf-8')
        at_index = encoded_email.rfind(b'@')
        email_length = len(encoded_email)
        if not (1 <= at_index <= EMAIL_LOCAL_PART_MAX and at_index + 1 < email_length <= EMAIL_MAX):
            raise InvalidParamsException(error_msg)


def _validate_country(country_code: str):
    if country_code.upper() not in iso3166.countries_by_alpha3:
        raise InvalidParamsException("Invalid alpha-3 country code")


def validate_irep(context: 'IconScoreContext', irep: int, prep: 'PRep'):
    term: 'Term' = context.engine.prep.term
    _validate_irep(irep=irep,
                   prev_irep=prep.irep,
                   prev_irep_block_height=prep.irep_block_height,
                   term_start_block_height=term.start_block_height,
                   term_total_supply=term.total_supply,
                   main_prep_count=context.main_prep_count)


def validate_np_irep(context: 'IconScoreContext', irep: int):
    term: 'Term' = context.engine.prep.term
    _validate_irep(irep=irep,
                   prev_irep=term.irep,
                   prev_irep_block_height=0,
                   term_start_block_height=1,
                   term_total_supply=term.total_supply,
                   main_prep_count=context.main_prep_count)


def _validate_irep(irep: int,
                   prev_irep: int,
                   prev_irep_block_height: int,
                   term_start_block_height: int,
                   term_total_supply: int,
                   main_prep_count: int):

    """
    beta1 + beta2 <= total_supply * IISS_MAX_IREP_PERCENTAGE / 100
    = (1/2 * irep * MAIN_PREP_COUNT + 1/2 * irep * PERCENTAGE_FOR_BETA_2) * IISS_MONTH
    = irep * (MAIN_PREP_COUNT + PERCENTAGE_FOR_BETA_2) * IISS_MONTH / 2
    <= total_supply * IISS_MAX_IREP_PERCENTAGE / 100

    irep <= total_supply * IISS_MAX_IREP_PERCENTAGE / (IISS_MONTH / 2 * 100 * (MAIN_PREP_COUNT + PERCENTAGE_FOR_BETA_2)
    """

    if prev_irep_block_height >= term_start_block_height:
        raise InvalidRequestException("I-Rep can be changed only once during a term")

    min_irep: int = max(prev_irep * 8 // 10, IISS_MIN_IREP)  # 80% of previous irep

    maximum_calculated_irep: int = \
        term_total_supply * IISS_MAX_IREP_PERCENTAGE // \
        (IISS_MONTH // 2 * 100 * (main_prep_count + PERCENTAGE_FOR_BETA_2))

    max_irep: int = min(prev_irep * 12 // 10, maximum_calculated_irep)  # 120% of previous irep

    if not min_irep <= irep <= max_irep:
        raise InvalidParamsException(f"I-Rep({irep}) out of range: {min_irep:,} ~ {max_irep:,}")



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..database.db import ContextDatabase
    from ..iconscore.icon_score_context import IconScoreContext


class EngineBase(ABC):
    def __init__(self):
        pass

    def open(self, *args, **kwargs):
        pass

    def close(self):
        pass

    def rollback(self, context: 'IconScoreContext', block_height: int, block_hash: bytes):
        pass


class StorageBase(ABC):

    def __init__(self, db: 'ContextDatabase'):
        """Constructor

        :param db: (Database) state db wrapper
        """
        self._db = db

    def open(self, *args, **kwargs):
        pass

    def close(self, context: 'IconScoreContext'):
        """Close the embedded database.

        :param context:
        """
        if self._db:
            self._db.close(context)
            self._db = None

    def rollback(self, context: 'IconScoreContext', block_height: int, block_hash: bytes):
        pass



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .message import *


class MessageUnpacker(object):
    def __init__(self):
        self._unpacker = msgpack.Unpacker(raw=True)
        self._iter = None
        self._response_classes = {
            MessageType.VERSION: VersionResponse,
            MessageType.CLAIM: ClaimResponse,
            MessageType.QUERY: QueryResponse,
            MessageType.CALCULATE: CalculateResponse,
            MessageType.COMMIT_BLOCK: CommitBlockResponse,
            MessageType.COMMIT_CLAIM: CommitClaimResponse,
            MessageType.QUERY_CALCULATE_STATUS: QueryCalculateStatusResponse,
            MessageType.QUERY_CALCULATE_RESULT: QueryCalculateResultResponse,
            MessageType.INIT: InitResponse,
            MessageType.START_BLOCK: StartBlockResponse,
            MessageType.READY: ReadyNotification,
            MessageType.CALCULATE_DONE: CalculateDoneNotification,
            MessageType.ROLLBACK: RollbackResponse,
        }

    def feed(self, data: bytes):
        self._unpacker.feed(data)

    def __iter__(self):
        self._iter = iter(self._unpacker)
        return self

    def __next__(self):
        try:
            message: list = next(self._iter)
            return self._parse(message)
        except:
            self._iter = None
            raise

    def _parse(self, message: list):
        msg_type: MessageType = MessageType(message[0])
        response_class = self._response_classes[msg_type]
        return response_class.from_list(message)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util
import os

from ..base.exception import IllegalFormatException

CODE_ATTR = 'co_code'
CODE_NAMES_ATTR = 'co_names'

BLACKLIST_RESERVED_KEYWORD = ['exec', 'eval', 'compile', '__import__']

LOAD_CONST = 100
IMPORT_STAR = 84
IMPORT_NAME = 108
IMPORT_FROM = 109

BASE_PACKAGE = 'iconservice'

# == OPCODE ==
# 20 LOAD_CONST(key1) value1
# 22 LOAD_CONST(key2) value2
# ============
# ...
OPCODE_HEADER_END_INDEX = 4


class ScorePackageValidator(object):
    WHITELIST_IMPORT = {}
    CUSTOM_IMPORT_LIST = []
    ICONSERVICE_WHITELIST = []

    @classmethod
    def _init_iconservice_whitelist(cls):
        if cls.ICONSERVICE_WHITELIST:
            return
        else:
            cls._load_iconservice_whitelist()

    @classmethod
    def _load_iconservice_whitelist(cls):
        spec = importlib.util.find_spec(BASE_PACKAGE)
        code = spec.loader.get_code(BASE_PACKAGE)

        if not hasattr(code, CODE_ATTR):
            return

        byte_code_list = [x for x in code.co_code]

        length_byte_code_list = len(byte_code_list)
        for code_index in range(OPCODE_HEADER_END_INDEX, length_byte_code_list, 2):
            key = byte_code_list[code_index]
            if IMPORT_NAME == key:
                # import_name_index = byte_code_list[code_index + 1]
                from_list_index = byte_code_list[code_index - 1]
                level_index = byte_code_list[code_index - 3]

                level = code.co_consts[level_index]
                if level == 0:
                    continue

                from_list = code.co_consts[from_list_index]
                cls.ICONSERVICE_WHITELIST.extend(from_list)
        return cls.ICONSERVICE_WHITELIST

    @classmethod
    def execute(cls,
                whitelist_table: dict,
                pkg_root_path: str,
                pkg_root_package: str) -> callable:

        cls.WHITELIST_IMPORT = whitelist_table
        cls.CUSTOM_IMPORT_LIST = cls._make_custom_import_list(pkg_root_path)
        cls._init_iconservice_whitelist()

        # in order for the new module to be noticed by the import system
        importlib.invalidate_caches()

        for imp in cls.CUSTOM_IMPORT_LIST:
            full_name = f'{pkg_root_package}.{imp}'

            spec = importlib.util.find_spec(full_name)
            code = spec.loader.get_code(full_name)

            cls._validate_import_from_code(code)
            cls._validate_import_from_const(code.co_consts)
            cls._validate_blacklist_keyword_from_names(code.co_names)

    @classmethod
    def _make_custom_import_list(cls,
                                 pkg_root_path: str) -> list:
        tmp_list = []
        for dirpath, _, filenames in os.walk(pkg_root_path):
            for file in filenames:
                file_name, extension = os.path.splitext(file)
                if extension != '.py':
                    continue
                sub_pkg_path = os.path.relpath(dirpath, pkg_root_path)
                if sub_pkg_path == '.':
                    pkg_path = file_name
                else:
                    # sub_package
                    sub_pkg_path = sub_pkg_path.replace('/', '.')
                    pkg_path = f'{sub_pkg_path}.{file_name}'
                tmp_list.append(pkg_path)
        return tmp_list

    @classmethod
    def _validate_blacklist_keyword_from_names(cls,
                                               co_names: tuple):
        for co_name in co_names:
            if co_name in BLACKLIST_RESERVED_KEYWORD:
                raise IllegalFormatException(f'Blacklist keyword found: {co_name}')

    @classmethod
    def _validate_import_from_code(cls,
                                   code):
        if not hasattr(code, CODE_ATTR):
            return

        byte_code_list = [x for x in code.co_code]

        length_byte_code_list = len(byte_code_list)
        for code_index in range(OPCODE_HEADER_END_INDEX, length_byte_code_list, 2):
            key = byte_code_list[code_index]
            if IMPORT_NAME == key:
                cls._validate_import(code_index, byte_code_list, code.co_names, code.co_consts)

    @classmethod
    def _validate_import_from_const(cls,
                                    co_consts: tuple):
        for co_const in co_consts:
            if not hasattr(co_const, CODE_ATTR):
                continue
            cls._validate_import_from_code(co_const)
            cls._validate_import_from_const(co_const.co_consts)
            if hasattr(co_const, CODE_NAMES_ATTR):
                cls._validate_blacklist_keyword_from_names(co_const.co_names)

    @classmethod
    def _validate_import(cls,
                         current_index: int,
                         byte_code_list: list,
                         co_names: tuple,
                         co_consts: tuple):
        """ example
        20 LOAD_CONST               0 (0)
        22 LOAD_CONST               3 (('pack', 'unpack', 'iter_unpack'))
        24 IMPORT_NAME              1 (struct)
        26 IMPORT_FROM              3 (pack)
        28 STORE_NAME               3 (pack)
        30 IMPORT_FROM              4 (unpack)
        32 STORE_NAME               4 (unpack)
        34 IMPORT_FROM              5 (iter_unpack)
        36 STORE_NAME               5 (iter_unpack)

        6 LOAD_CONST                0 (0)
        18 LOAD_CONST               2 (None)
        20 IMPORT_NAME              2 (os)
        22 STORE_NAME               2 (os)
        24 LOAD_CONST               0 (0)
        26 LOAD_CONST               2 (None)
        28 IMPORT_NAME              3 (json)
        30 STORE_NAME               3 (json)
        """

        from_list_op_code_key = byte_code_list[current_index - 2]
        level_op_code_key = byte_code_list[current_index - 4]
        if LOAD_CONST != from_list_op_code_key or LOAD_CONST != level_op_code_key:
            raise IllegalFormatException('Invalid import opcode')

        import_name_index = byte_code_list[current_index + 1]
        from_list_index = byte_code_list[current_index - 1]
        level_index = byte_code_list[current_index - 3]

        import_name = co_names[import_name_index]
        from_list = co_consts[from_list_index]
        level = co_consts[level_index]

        if level > 0:
            return

        if import_name not in cls.WHITELIST_IMPORT:
            raise IllegalFormatException(f'Invalid import name: {import_name}')

        if from_list is None:
            # only using import
            return
        else:
            next_op_code_key = byte_code_list[current_index + 2]
            if IMPORT_STAR == next_op_code_key:
                # import_star
                if from_list[0] != '*':
                    raise IllegalFormatException(f'Invalid star import: {import_name}')
            elif IMPORT_FROM == next_op_code_key:
                # import from
                for import_from in from_list:
                    if '*' not in cls.WHITELIST_IMPORT[import_name] and \
                            import_from not in cls.WHITELIST_IMPORT[import_name]:
                        raise IllegalFormatException(f'Invalid import name: {import_name}')
                    elif import_name in BASE_PACKAGE and \
                            import_from not in cls.ICONSERVICE_WHITELIST:
                        raise IllegalFormatException(f'Invalid permission import: {import_from}')

            else:
                raise IllegalFormatException('Invalid import opcode')



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import threading
from typing import Optional, List, TYPE_CHECKING

from ...base.exception import FatalException

if TYPE_CHECKING:
    from ..icon_score_context import IconScoreContext

_thread_local_data = threading.local()


class ContextContainer(object):
    """ContextContainer mixin

    Every class which inherits ContextContainer can share IconScoreContext instance
    in the current thread.
    """

    @staticmethod
    def _get_context() -> Optional['IconScoreContext']:
        context_stack: List['IconScoreContext'] = getattr(_thread_local_data, 'context_stack', None)

        if context_stack is not None and len(context_stack) > 0:
            return context_stack[-1]
        else:
            return None

    @staticmethod
    def _push_context(context: 'IconScoreContext') -> None:
        context_stack: List['IconScoreContext'] = getattr(_thread_local_data, 'context_stack', None)

        if context_stack is None:
            context_stack = []
            setattr(_thread_local_data, 'context_stack', context_stack)

        context_stack.append(context)

    @staticmethod
    def _pop_context() -> 'IconScoreContext':
        """Delete the last pushed context of the current thread
        """
        context_stack: List['IconScoreContext'] = getattr(_thread_local_data, 'context_stack', None)

        if context_stack is not None and len(context_stack) > 0:
            return context_stack.pop()
        else:
            raise FatalException('Failed to pop a context out of context_stack')

    @staticmethod
    def _clear_context() -> None:
        setattr(_thread_local_data, 'context_stack', None)

    @staticmethod
    def _get_context_stack_size() -> int:
        context_stack: List['IconScoreContext'] = getattr(_thread_local_data, 'context_stack', None)
        return 0 if context_stack is None else len(context_stack)


class ContextGetter(object):
    """The class which refers to IconScoreContext should inherit ContextGetter
    """

    @property
    def _context(self) -> 'IconScoreContext':
        return ContextContainer._get_context()



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import asyncio
import copy
import os
import subprocess
import sys
from enum import IntEnum
from typing import TYPE_CHECKING

from iconcommons.icon_config import IconConfig
from iconcommons.logger import Logger

from iconservice.icon_config import default_icon_config
from iconservice.icon_constant import ICON_SCORE_QUEUE_NAME_FORMAT, ICON_SERVICE_PROCTITLE_FORMAT, ConfigKey
from .__version__ import __version__

if TYPE_CHECKING:
    from .icon_inner_service import IconScoreInnerStub

_TAG = "CLI"


class ExitCode(IntEnum):
    SUCCEEDED = 0
    INVALID_COMMAND = 1


def main():
    parser = argparse.ArgumentParser(prog='icon_service_cli.py', usage=f"""
    ==========================
    iconservice {__version__}
    ==========================
    iconservice commands:
        start : iconservice start
        stop : iconservice stop

        -c : json configure file path
        -sc : icon score root path ex).score
        -st : icon score state db root path ex).state
        -at : amqp target info [IP]:[PORT]
        -ak : key sharing peer group using queue name. use it if one more peers connect one MQ
        -ch : loopchain channel ex) loopchain_default
        -fg : foreground process
        -tbears : tbears mode
    """)

    parser.add_argument('command', type=str,
                        nargs='*',
                        choices=['start', 'stop'],
                        help='iconservice type [start|stop]')
    parser.add_argument("-sc", dest=ConfigKey.SCORE_ROOT_PATH, type=str, default=None,
                        help="icon score root path  example : .score")
    parser.add_argument("-st", dest=ConfigKey.STATE_DB_ROOT_PATH, type=str, default=None,
                        help="icon score state db root path  example : .statedb")
    parser.add_argument("-ch", dest=ConfigKey.CHANNEL, type=str, default=None,
                        help="icon score channel")
    parser.add_argument("-ak", dest=ConfigKey.AMQP_KEY, type=str, default=None,
                        help="icon score amqp_key : [amqp_key]")
    parser.add_argument("-at", dest=ConfigKey.AMQP_TARGET, type=str, default=None,
                        help="icon score amqp_target : [127.0.0.1]")
    parser.add_argument("-c", dest=ConfigKey.CONFIG, type=str, default=None,
                        help="icon score config")
    parser.add_argument("-fg", dest='foreground', action='store_true',
                        help="icon score service run foreground")
    parser.add_argument("-tbears", dest=ConfigKey.TBEARS_MODE, action='store_true',
                        help="tbears mode")
    parser.add_argument("-steptrace", dest=ConfigKey.STEP_TRACE_FLAG, action="store_true", help="enable step tracing")

    args = parser.parse_args()

    if len(args.command) < 1:
        parser.print_help()
        sys.exit(ExitCode.INVALID_COMMAND.value)

    conf_path = args.config

    if conf_path is not None:
        if not IconConfig.valid_conf_path(conf_path):
            print(f'invalid config file : {conf_path}')
            sys.exit(ExitCode.INVALID_COMMAND.value)
    if conf_path is None:
        conf_path = str()

    conf = IconConfig(conf_path, copy.deepcopy(default_icon_config))
    conf.load()
    conf.update_conf(dict(vars(args)))
    Logger.load_config(conf)

    command = args.command[0]
    if command == 'start' and len(args.command) == 1:
        result = _start(conf)
    elif command == 'stop' and len(args.command) == 1:
        result = _stop(conf)
    else:
        parser.print_help()
        result = ExitCode.INVALID_COMMAND.value
    sys.exit(result)


def _start(conf: 'IconConfig') -> int:
    if not _check_if_process_running(conf):
        _start_process(conf)
    Logger.info(f'start_command done!', _TAG)
    return ExitCode.SUCCEEDED


def _stop(conf: 'IconConfig') -> int:
    async def __stop():
        await stop_process(conf)

    if _check_if_process_running(conf):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(__stop())

    Logger.info(f'stop_command done!', _TAG)
    return ExitCode.SUCCEEDED


def _start_process(conf: 'IconConfig'):
    Logger.info('start_server() start')
    python_module_string = 'iconservice.icon_service'

    converted_params = {'-sc': conf[ConfigKey.SCORE_ROOT_PATH],
                        '-st': conf[ConfigKey.STATE_DB_ROOT_PATH],
                        '-ch': conf[ConfigKey.CHANNEL], '-ak': conf[ConfigKey.AMQP_KEY],
                        '-at': conf[ConfigKey.AMQP_TARGET], '-c': conf.get(ConfigKey.CONFIG)}

    custom_argv = []
    for k, v in converted_params.items():
        if v is None:
            continue
        custom_argv.append(k)
        custom_argv.append(str(v))
    if conf[ConfigKey.TBEARS_MODE]:
        custom_argv.append('-tbears')
    if conf[ConfigKey.STEP_TRACE_FLAG]:
        custom_argv.append('-steptrace')

    is_foreground = conf.get('foreground', False)
    if is_foreground:
        from iconservice.icon_service import run_in_foreground
        del conf['foreground']
        run_in_foreground(conf)
    else:
        subprocess.Popen([sys.executable, '-m', python_module_string, *custom_argv], close_fds=True)
    Logger.info('start_process() end')


async def stop_process(conf: 'IconConfig'):
    icon_score_queue_name = _make_icon_score_queue_name(conf[ConfigKey.CHANNEL], conf[ConfigKey.AMQP_KEY])
    stub = await _create_icon_score_stub(conf[ConfigKey.AMQP_TARGET], icon_score_queue_name)
    await stub.async_task().close()
    Logger.info(f'stop_process_icon_service!', _TAG)


def _check_if_process_running(conf: 'IconConfig') -> bool:
    proc_title = ICON_SERVICE_PROCTITLE_FORMAT.format(**
                                                      {ConfigKey.SCORE_ROOT_PATH: conf[ConfigKey.SCORE_ROOT_PATH],
                                                       ConfigKey.STATE_DB_ROOT_PATH: conf[ConfigKey.STATE_DB_ROOT_PATH],
                                                       ConfigKey.CHANNEL: conf[ConfigKey.CHANNEL],
                                                       ConfigKey.AMQP_KEY: conf[ConfigKey.AMQP_KEY],
                                                       ConfigKey.AMQP_TARGET: conf[ConfigKey.AMQP_TARGET]})
    cmd_lines = _get_process_command_list(b'icon_service.')
    if cmd_lines:
        for cmdline in cmd_lines:
            if cmdline == proc_title:
                return True
    return False


def _get_process_command_list(prefix: bytes) -> list:
    if os.path.exists('/proc'):
        cmd_lines = []
        pids = [pid for pid in os.listdir('/proc') if pid.isdigit()]
        for pid in pids:
            try:
                cmdpath = os.path.join('/proc', pid, 'cmdline')
                cmdline = open(cmdpath, 'rb').read().rstrip(b'\x00')
                if cmdline.startswith(prefix):
                    cmd_lines.append(cmdline.decode())
            except IOError:
                continue
    else:
        result = subprocess.run(['ps', '-eo', 'command'], stdout=subprocess.PIPE)
        cmd_lines = [cmdline.decode().rstrip()
                     for cmdline in result.stdout.split(b'\n')
                     if cmdline.startswith(prefix)]
    return cmd_lines


def _make_icon_score_queue_name(channel: str, amqp_key: str) -> str:
    return ICON_SCORE_QUEUE_NAME_FORMAT.format(channel_name=channel, amqp_key=amqp_key)


async def _create_icon_score_stub(amqp_target: str, icon_score_queue_name: str) -> 'IconScoreInnerStub':
    from .icon_inner_service import IconScoreInnerStub

    stub = IconScoreInnerStub(amqp_target, icon_score_queue_name)
    await stub.connect()
    return stub
